'''The :mod:`glyspace` module provides an API for interacting with `GlyTouCan <https://glytoucan.org/>`_
and `UnicarbKB <http://www.unicarbkb.org/>`_. The interaction may be done by executing prepared
:term:`SPARQL` queries through the provided methods, executing user-provided :term:`SPARQL` queries,
:term:`RDF` Graph-operation supported by `RDFLib <https://rdflib.readthedocs.io/en/stable/>`_,
or using :term:`RDF`-object mapping via :meth:`GlySpaceRDFClient.get` and related methods.

'''

import logging
import re
import warnings

from numbers import Number
from collections import defaultdict, OrderedDict
from six import text_type

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote

import requests
with warnings.catch_warnings():
    from rdflib import ConjunctiveGraph, Namespace, URIRef, Literal, Graph
    from rdflib.namespace import split_uri
    from glypy.io import glycoct, iupac, wurcs, _glycordf

# http://glytoucan.org/glyspace/documentation/apidoc.html
# http://code.glytoucan.org/system/glyspace

logger = logging.getLogger(__name__)

ssl_verification = False


sparql_endpoint = "http://ts.glytoucan.org/sparql"

_get_all_monosaccharides_sparql = r'''
PREFIX glycan: <http://purl.jp/bio/12/glyco/glycan#>
PREFIX glytoucan:  <http://www.glytoucan.org/glyco/owl/glytoucan#>

SELECT DISTINCT ?Saccharide ?Sequence
WHERE {
    ?Saccharide glycan:has_glycosequence ?GlycoSequence .
    ?GlycoSequence glycan:has_sequence ?Sequence .
    ?GlycoSequence glycan:in_carbohydrate_format glycan:carbohydrate_format_glycoct .
    ?Saccharide a glycan:monosaccharide
}
'''


_get_all_glycans_sparql = r'''
PREFIX glycan: <http://purl.jp/bio/12/glyco/glycan#>
PREFIX glytoucan:  <http://www.glytoucan.org/glyco/owl/glytoucan#>


SELECT DISTINCT ?Saccharide ?PrimaryId ?Sequence
WHERE {
    ?Saccharide glytoucan:has_primary_id ?PrimaryId .
    ?Saccharide glycan:has_glycosequence ?GlycoSequence .
    ?GlycoSequence glycan:has_sequence ?Sequence .
    ?GlycoSequence glycan:in_carbohydrate_format glycan:carbohydrate_format_glycoct
}
ORDER BY ?PrimaryId
'''

NSGlycan = _glycordf.NSGlycan
NSGlyTouCan = Namespace("http://www.glytoucan.org/glyco/owl/glytoucan#")
NSGlycoinfo = Namespace("http://rdf.glycoinfo.org/glycan/")
NSGlycomeDB = Namespace("http://rdf.glycome-db.org/glycan/")
NSSKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
NSUniprotCore = Namespace("http://purl.uniprot.org/core/")
NSUniprotEntity = Namespace("http://purl.uniprot.org/uniprot/")
NSTaxonomy = Namespace("http://purl.uniprot.org/taxonomy/")
NSDCTerms = Namespace("http://purl.org/dc/terms/")
NSGlycoconjugate = Namespace("http://purl.jp/bio/12/glyco/conjugate#")
NSUnicarbKB = Namespace("http://rdf.unicarbkb.org/")
NSUnicarbKBStructure = Namespace("http://rdf.unicarbkb.org/structure/")
NSUnicarbKBProtein = Namespace("http://unicarbkb.org/unicarbkbprotein/")
NSFaldo = Namespace("http://www.biohackathon.org/resource/faldo/")
NSPato = Namespace("http://purl.obolibrary.org/obo/uo.owl")
NSUnicorn = Namespace("http://purl.jp/bio/12/glyco/unicorn/")
NSOwl = Namespace("http://www.w3.org/2002/07/owl#")
NSSIO = Namespace("http://semanticscience.org/resource/")
NSFOAF = Namespace("http://xmlns.com/foaf/0.1/")
NSRDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")

GlycoRDF = _glycordf.glycordf

_uniprot_taxon_uri = u"http://www.uniprot.org/taxonomy/{}.rdf"


def _camel_to_snake(name):
    '''
    Many properties and predicate names are in camelCase, which is
    not consistent with Python naming conventions.
    '''
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', re.sub(r'(.)([A-Z][a-z]+)', r"\1_\2", name)).lower()


def _uri_to_identifier(uri):
    name = _camel_to_snake(split_uri(uri)[1])
    return name


class LRUDict(object):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        maxsize = kwargs.pop("maxsize", 24)
        self.store = OrderedDict()
        self.maxsize = maxsize
        self.purge()

    def __len__(self):
        return len(self.store)

    def popitem(self, last=True):
        return self.store.popitem(last=last)

    def pop(self, key, default=None):
        return self.store.pop(key, default)

    def purge(self):
        overflow = max(0, len(self) - self.maxsize)
        for _ in range(overflow):
            self.popitem(last=False)

    def __repr__(self):
        return "LRUDict(%r)" % (dict(self.store),)

    def __contains__(self, key):
        return key in self.store

    def __iter__(self):
        return iter(self.store)

    def keys(self):
        return self.store.keys()

    def values(self):
        return self.store.values()

    def items(self):
        return self.store.items()

    def __getitem__(self, key):
        value = self.store[key]
        self._mark_used(key)
        return value

    def __setitem__(self, key, value):
        self.store[key] = value
        self.purge()

    def _mark_used(self, key):
        value = self.store.pop(key, None)
        self.store[key] = value


class PredicateDescriptor(text_type):

    """A specialization of the unicode text type for representing a string which
    is a convenient suffix of a complete URI denoting a predicate.

    This type is used to attach a more verbose URI to a short string used for a
    :class:`ReferenceEntity`'s attributes.

    This class should be immutable, and caches all unique URIs.
    """

    __slots__ = ("source", )
    _cache = {}

    @classmethod
    def bind(cls, uri):
        return cls(_uri_to_identifier(uri), uri)

    def __new__(cls, value, source):
        if source in cls._cache:
            return cls._cache[source]
        named = super(PredicateDescriptor, cls).__new__(cls, value)
        named.source = source
        cls._cache[source] = named
        return named


class ReferenceEntity(object):
    '''
    A ReferenceEntity is a generic type to for storing results from
    a semantic query. Its attribute names are usually derived from
    predicates, and their values may be scalar or lists. It is usually
    constructed from :meth:`get` from a :class:`BoundURIRef` instance.

    A preprocessor may add new attributes to a ReferenceEntity during
    construction, such as `structure_` with a |Glycan| instance value
    when the referenced entity satisfies `glycan:has_glycosequence`.
    '''

    query_type = 'subject'

    def __init__(self, uriref, **kwargs):
        self._uriref = uriref
        for k, v in kwargs.items():
            setattr(self, k, v)

        self._keys = {k for k in kwargs.keys() if isinstance(k, PredicateDescriptor)}
        self._object_of = None

    def __iter__(self):
        for k in self._keys:
            yield k, getattr(self, k)

    def __getitem__(self, k):  # pragma: no cover
        return getattr(self, k)

    def __setitem__(self, k, v):  # pragma: no cover
        setattr(self, k, v)
        if isinstance(k, PredicateDescriptor):
            self._keys.add(k)

    def __repr__(self):  # pragma: no cover
        tokens = [self.__class__.__name__]
        for k, v in self:
            if isinstance(v, list):
                v_str = ',\n    '.join(map(str, v[:20]))
                if len(v) > 20:
                    v_str += '\n    ...\n'
            else:
                v_str = v
            tokens.append("%s: %s" % (k, v_str))
        return "\n".join(tokens)

    def __root__(self):
        return self.structure_.__root__()

    def __tree__(self):
        return self.structure_

    @property
    def _graph(self):
        return self._uriref._graph

    @property
    def object_of(self):
        if self._object_of is None:
            self._object_of = self._graph.get(self._uriref, query_type='object')
        return self._object_of

    @property
    def uriref(self):
        return self._uriref

    def to_graph(self, graph=None, deep=False, visited=None):
        if graph is None:
            graph = Graph()
            for prefix, ns in self._graph.namespaces():
                graph.bind(prefix, ns)

        if visited is None:
            visited = set()
        key = (self.uriref, self.query_type)
        if key in visited:
            return graph
        else:
            visited.add(key)

        for predicate, value in self:
            if isinstance(value, (text_type, Number)):
                is_uri = isinstance(value, URIRef)
                if not is_uri:
                    value_n = Literal(value)
                else:
                    value_n = URIRef(value)
                graph.add((self.uriref, URIRef(predicate.source), value_n))
                if is_uri and deep:
                    try:
                        value.to_graph(graph, deep=deep, visited=visited)
                    except AttributeError:
                        self._graph.get(value).to_graph(graph, deep=deep, visited=visited)
            elif isinstance(value, list):
                value_collection = value
                for value in value_collection:
                    if isinstance(value, (text_type, Number)):
                        is_uri = isinstance(value, URIRef)
                        if not is_uri:
                            value_n = Literal(value)
                        else:
                            value_n = URIRef(value)
                        graph.add((self.uriref, URIRef(predicate.source), value_n))
                        if is_uri and deep:
                            try:
                                value.to_graph(graph, deep=deep, visited=visited)
                            except AttributeError:
                                self._graph.get(value).to_graph(graph, deep=deep, visited=visited)
                    else:
                        raise TypeError(
                            "Could not translate %r of type %r for predicate %s" % (
                                value, type(value), predicate.source))
            else:
                raise TypeError(
                    "Could not translate %r of type %r for predicate %s" % (
                        value, type(value), predicate.source))

        if deep:
            self.object_of.to_graph(graph, deep=deep, visited=visited)
        return graph


class PredicateCollection(ReferenceEntity):
    query_type = 'predicate'

    def to_graph(self, graph=None, deep=False, visited=None):
        if graph is None:
            graph = Graph()
            for prefix, ns in self._graph.namespaces():
                graph.bind(prefix, ns)
        if visited is None:
            visited = set()
        key = (self.uriref, self.query_type)
        if key in visited:
            return graph
        else:
            visited.add(key)
        for subj, obj in self:
            if isinstance(subj, (unicode, str, int, float)):
                is_uri_subj = isinstance(subj, URIRef)
                if not is_uri_subj:
                    subj = Literal(subj)
            if isinstance(obj, (unicode, str, int, float)):
                is_uri_obj = isinstance(obj, URIRef)
                if not is_uri_obj:
                    obj = Literal(obj)
            graph.add((subj, self.uriref, obj))
            if is_uri_subj and deep:
                try:
                    subj.to_graph(graph, deep=deep, visited=visited)
                except AttributeError:
                    self._graph.get(subj).to_graph(graph, deep=deep, visited=visited)
            if is_uri_obj:
                try:
                    obj.to_graph(graph, deep=deep, visited=visited)
                except AttributeError:
                    self._graph.get(obj).to_graph(graph, deep=deep, visited=visited)

        return graph


class ObjectOfEntity(ReferenceEntity):
    query_type = 'object'

    def to_graph(self, graph=None, deep=False, visited=None):
        if graph is None:
            graph = Graph()
            for prefix, ns in self._graph.namespaces():
                graph.bind(prefix, ns)

        if visited is None:
            visited = set()
        key = (self.uriref, self.query_type)
        if key in visited:
            return graph
        else:
            visited.add(key)

        for predicate, subj in self:
            if isinstance(subj, (unicode, str, int, float)):
                is_uri = isinstance(subj, URIRef)
                if not is_uri:
                    subj = Literal(subj)
                graph.add((subj, URIRef(predicate.source), self.uriref))
                if is_uri and deep:
                    try:
                        subj.to_graph(graph, deep=deep, visited=visited)
                    except AttributeError:
                        self._graph.get(subj).to_graph(graph, deep=deep, visited=visited)
            elif isinstance(subj, list):
                subj_collection = subj
                for subj in subj_collection:
                    if isinstance(subj, (unicode, str, int, float)):
                        is_uri = isinstance(subj, URIRef)
                        if not is_uri:
                            subj = Literal(subj)
                        graph.add((subj, URIRef(predicate.source), self.uriref))
                        if is_uri and deep:
                            try:
                                subj.to_graph(graph, deep=deep, visited=visited)
                            except AttributeError:
                                self._graph.get(subj).to_graph(graph, deep=deep, visited=visited)
                    else:
                        raise TypeError(
                            "Could not translate %r of type %r for predicate %s" % (
                                subj, type(subj), predicate.source))
            else:
                raise TypeError(
                    "Could not translate %r of type %r for predicate %s" % (
                        subj, type(subj), predicate.source))
        return graph


class BoundURIRef(URIRef):
    '''
    A subclass of :class:`rdflib.term.URIRef` which bakes in a way to
    fetch the referenced subgraph (in the semantic web sense) as a :class:`.ReferenceEntity` by keeping
    a reference (in the memory address sense) to the :class:`GlySpaceRDFClient` instance
    which fetched the object it is attached to.

    It has some convenience features for interactive use such as implict
    resource fetching when checking for attribute completions.

    Attributes
    ----------
    _graph: :class:`RDFClientBase`
    _result_ref: :class:`ReferenceEntity`
        A reference to the ReferenceEntity fetched from this URI.
        Acts as a cache

    '''
    __slots__ = ("_graph", "_result_ref")

    def __new__(cls, value, base=None, source=None):
        rt = URIRef.__new__(cls, value, base)
        rt._graph = source
        rt._result_ref = None
        return rt

    def __hash__(self):
        return URIRef.__hash__(self)

    def get(self, simplify=True, refresh=False, query_type='auto'):
        """Get the referenced entity either from :attr:`_graph`
        or the cached reference in :attr:`_result_ref`

        Parameters
        ----------
        simplify : bool, optional
            As in :meth:`RDFClientBase.get`
        refresh : bool, optional
            If `True`, always request the URI's semantic reference,
            ignoring :attr:`_result_ref`

        Returns
        -------
        ReferenceEntity
        """
        if self._result_ref is None or refresh:
            result = self._graph.get(self, simplify=simplify, query_type=query_type)
            self._result_ref = result
            return result
        else:
            return self._result_ref

    __call__ = get

    def _repr_pretty_(self, p, cycle):
        return p.pretty(str(self))

    def __getattr__(self, name):
        '''
        A convenience method to forward missed attribute lookup
        to the referenced entity. Calls :meth:`__call__`, which
        may initiate a network request.
        '''
        # IPython's displayhook tries to find the attribute _ipython_display_
        # which naturally is not found and initiates a network request. This
        # causes undesirable lag when using the interactive terminal, so short-
        # circuit.
        if name.startswith("_ipython"):
            raise AttributeError()
        reference = self()
        try:
            return getattr(reference, name)
        except AttributeError:
            raise AttributeError("Neither %s nor its reference have an attribute %s" % (self, name))

    def __dir__(self):  # pragma: no cover
        if self._result_ref is None:
            self()
        return dir(self._result_ref)

    def __eq__(self, other):
        '''
        Overrides the equality method of :class:`rdflib.term.URIRef` which
        does exact :func:`type` comparison before comparing contents to short-circuit
        on non-URIs.
        '''
        result = str(self) == str(other)
        return result

    def n3(self, namespace_manager=None):
        try:
            text = super(BoundURIRef, self).n3(namespace_manager)
            return text
        except Exception:
            if namespace_manager:
                return namespace_manager.normalizeUri(self)
            else:
                return "<%s>" % self


class ChainFunctionDict(defaultdict):
    '''
    A wrapper around :class:`defaultdict(list)` which
    keys on :class:`rdflib.term.URIRef` strings. Added values
    should be callables which will be invoked in the order given on
    an entity state `dict` and each `object` which is linked by the key `predicate`.
    '''
    def __init__(self, *args, **kwargs):
        defaultdict.__init__(self, list, *args, **kwargs)

    def __call__(self, predicate, state, obj):
        ret_val_was_none = False
        non_none_ret_val = None
        for f in self[predicate]:
            ret_val = f(state, obj)
            if ret_val is not None:
                non_none_ret_val = ret_val
            else:
                ret_val_was_none = True
        if ret_val_was_none:
            return None
        else:
            return non_none_ret_val


class RDFClientBase(ConjunctiveGraph):
    _predicates_seen = set()
    predicate_processor_map = ChainFunctionDict()

    @classmethod
    def register_predicate_processor(cls, predicate):
        """Decorator to register a callable processor for a `URIRef` `predicate` with this type's
        :attr:`predicate_processor_map`. The actual decorated callable is returned unchanged.

        Parameters
        ----------
        predicate : rdflib.term.URIRef or str
            The type of URI to add the decorated callable
            to the processor chain of

        Returns
        -------
        callable
        """
        def wrapper(f):
            cls.predicate_processor_map[predicate].append(f)
            return f
        return wrapper

    def __init__(self, sparql_endpoint, accession_ns, cache_size=100):
        super(RDFClientBase, self).__init__(store="SPARQLStore")
        self.open(sparql_endpoint)
        self.accession_ns = accession_ns
        self.cache = LRUDict(maxsize=cache_size)

    def accession_to_uriref(self, accession):
        """Utility method to translate free strings into full URIs
        derived from this instance's :attr:`accession_ns`

        Parameters
        ----------
        accession : str
            A regular string comprised of just the accession number of
            an entity.

        Returns
        -------
        rdflib.term.URIRef
        """
        return self.accession_ns[str(accession)]

    def get(self, uriref, simplify=True, query_type='auto'):
        """Download all related information for `uriref` from the remote
        data source.

        Collects all the triples from the remote data source where `uriref` is
        the subject. If `uriref` is not the subject of any triples, it is re-queried
        as a predicate, storing the subject-object pairs.

        Any objects (and subjects) which are themselves :class:`rdflib.term.URIRef` instances
        will be converted into :class:`BoundURIRef` which will silently fetch the relevant
        entity from the remote source.

        If the predicate matches a processor rules, instead of it's object value being
        stored, the object will be transformed by each rule in the processor chain.

        Parameters
        ----------
        uriref: str or rdflib.term.URIRef
            A subject, predicate, or a database accession number to transform through
            :meth:`accession_to_uriref`
        simplify: bool, optional
            If true, any predicate with a single value will be a scalar,
            and any other will be a list.

        Returns
        -------
        ReferenceEntity
            An object representing the subject whose attributes are named after
            predicates with their objects as values.
        """
        if not isinstance(uriref, URIRef):
            uriref = self.accession_to_uriref(uriref)
        if (uriref, query_type) in self.cache:
            return self.cache[uriref, query_type]
        results = defaultdict(list)
        if query_type == 'auto' or query_type == 'subject':
            for subject, predicate, obj in set(self.triples((uriref, None, None))):
                predicate_name = PredicateDescriptor.bind(predicate)
                self._predicates_seen.add(predicate)
                if isinstance(obj, Literal):
                    obj = obj.toPython()
                elif isinstance(obj, URIRef):
                    obj = BoundURIRef(obj, source=self)
                if predicate in self.predicate_processor_map:
                    obj = self.predicate_processor_map(predicate, results, obj)
                if obj is not None:
                    results[predicate_name].append(obj)
            if results:
                query_type = 'subject'

        # If there were no results, the query might be a predicate, so try to find all the
        # pairs that satisfy it.
        if (len(results) == 0 and query_type == 'auto') or query_type == 'predicate':
            try:
                predicate_name = PredicateDescriptor.bind(uriref)
            except Exception:
                predicate_name = None
            if predicate_name is not None:
                for subject, predicate, obj in set(self.triples((None, uriref, None))):
                    if isinstance(obj, Literal):
                        obj = obj.toPython()
                    elif isinstance(obj, URIRef):
                        obj = BoundURIRef(obj, source=self)

                    if isinstance(subject, Literal):
                        subject = subject.toPython()
                    elif isinstance(subject, URIRef):
                        subject = BoundURIRef(subject, source=self)

                    results[predicate_name].append((subject, obj))
            if results:
                query_type = 'predicate'

        if query_type == 'object':
            for subject, predicate, obj in set(self.triples((None, None, uriref))):
                predicate_name = PredicateDescriptor.bind(predicate)
                self._predicates_seen.add(predicate)
                if isinstance(subject, Literal):
                    subject = subject.toPython()
                elif isinstance(subject, URIRef):
                    subject = BoundURIRef(subject, source=self)
                if predicate in self.predicate_processor_map:
                    subject = self.predicate_processor_map(predicate, results, subject)
                if subject is not None:
                    results[predicate_name].append(subject)

        if simplify:
            results = {k: v if len(v) > 1 else v[0] for k, v in results.items()}
        if query_type in ('subject', 'auto'):
            entity_type = ReferenceEntity
        elif query_type == 'predicate':
            entity_type = PredicateCollection
        elif query_type == 'object':
            entity_type = ObjectOfEntity
        else:
            raise ValueError("Could not determine query type %s" % (query_type,))
        results = entity_type(BoundURIRef(uriref, source=self), **results)
        self.cache[uriref, query_type] = results
        return results


class UniprotRDFClient(RDFClientBase):
    predicate_processor_map = ChainFunctionDict()
    _predicates_seen = set()
    _sparql_endpoint_uri = "http://sparql.uniprot.org/sparql/"

    def __init__(self):
        super(UniprotRDFClient, self).__init__(self._sparql_endpoint_uri, NSUniprotEntity)
        self.bind("up", NSUniprotCore)
        self.bind("taxonomy", NSTaxonomy)


class GlyTouCanRDFClient(RDFClientBase):
    '''
    An RDF Client for glySpace. The default namespace is `glycoinfo`, and
    the following namespaces are bound:

    .. code-block:: python

        glytoucan = NSGlyTouCan
        glycomedb = NSGlycomeDB
        glycan = NSGlycan
        glycoinfo = NSGlycoinfo
        skos = NSSKOS


    Attributes
    ----------
    predicate_processor_map: :class:`ChainFunctionDict`
        A dictionary keeping track of the chain of processors registered for
        each predicate.
    _sparql_endpoint_uri: str
        The web address to use as the remote backend for SPARQL queries.
        Passed on to :class:`rdflib.ConjunctiveGraph` and the `SPARQLStore`
        storage plugin.

    '''

    predicate_processor_map = ChainFunctionDict()
    _predicates_seen = set()
    _sparql_endpoint_uri = sparql_endpoint

    def __init__(self):
        super(GlyTouCanRDFClient, self).__init__(self._sparql_endpoint_uri, NSGlycoinfo)
        self.bind("glytoucan", NSGlyTouCan)
        self.bind("glycomedb", NSGlycomeDB)
        self.bind("glycan", NSGlycan)
        self.bind("glycoinfo", NSGlycoinfo)
        self.bind("skos", NSSKOS)
        self.bind("dcterms", NSDCTerms)

    def from_taxon(self, taxon, limit=None):
        r"""Fetch all accession numbers for all structures
        from the given taxonomic identifier, up to `limit` records.

        Equivalent to the following SPARQL

        .. code-block:: sparql

            SELECT DISTINCT ?saccharide WHERE {
                ?saccharide a glycan:saccharide .
                ?saccharide skos:exactMatch ?gdb .
                ?gdb glycan:has_reference ?ref .
                ?ref glycan:is_from_source ?source .
                ?source glycan:has_taxon ?taxon
                FILTER REGEX(str(?taxon), "http://www.uniprot.org/taxonomy/<taxonomy-accession>.rdf")
            }

        The REGEX filter is used because at current taxonomic information in Glycome-DB
        is encoded as a string instead of a URI.

        Parameters
        ----------
        taxon : str or int
            A string or number which corresponds to the taxonomy database
            id for the taxon of interest.
        limit : int, optional
            The maximum number of results to retrieve.

        Returns
        -------
        list of BoundURIRef
        """
        sparql = r'''
        SELECT DISTINCT ?accession ?taxon_id
        WHERE{
            ?saccharide  glytoucan:has_primary_id ?accession .
            ?saccharide glycan:is_from_source ?source.
            ?source a glycan:Source .
            VALUES ?taxon_id {"%s"}
            ?source dcterms:identifier ?taxon_id
        }
        '''
        query_string = sparql % (str(taxon), )
        if limit is not None:
            query_string += " limit %d" % limit
        results = self.query(query_string)
        k = results.vars[0]
        return [BoundURIRef(row[k], source=self) for row in results.bindings]

    def structures_with_motif(self, motif, limit=None):
        r"""Fetch all accession numbers and structures for all structures
        which contain the given motif accession, up to `limit` records.

        Equivalent to the following SPARQL

        .. code-block:: sparql

            SELECT DISTINCT ?saccharide WHERE {
                ?saccharide a glycan:saccharide .
                ?saccharide glycan:has_motif <motif-accession>
            }

        Parameters
        ----------
        motif : str
            The accession number of the motif of interest.
        limit : int, optional
            The maximum number of results to retrieve.

        Returns
        -------
        list of :class:`ReferenceEntity`
        """
        sparql = r'''
        SELECT DISTINCT ?saccharide WHERE {
                ?saccharide glycan:has_motif %s
        }
        '''
        if isinstance(motif, URIRef):
            motif_str = self.qname(motif)
        else:
            motif_str = self.qname(NSGlycoinfo[motif])
        query_string = sparql % motif_str
        if limit is not None:
            query_string += " limit %d" % limit
        results = self.query(query_string)
        k = results.vars[0]
        return [ReferenceEntity(row[k]) for row in results.bindings]

    def structure(self, *accessions):
        accumulator = []
        sparql = r'''
        SELECT DISTINCT ?saccharide ?glycoct WHERE {
            ?saccharide glycan:has_glycosequence ?sequence .
            FILTER CONTAINS(str(?sequence), "glycoct") .
            ?sequence glycan:has_sequence ?glycoct .
            FILTER ("%s" = str(?saccharide))
        }
        '''
        for accession in accessions:
            if isinstance(accession, URIRef):
                accession_str = str(accession)
            else:
                accession_str = str(NSGlycoinfo[accession])
            query_string = sparql % accession_str
            results = self.query(query_string)
            g = results.vars[1]
            glycoct_string = results.bindings[0][g]
            structure = glycoct.loads(glycoct_string)
            accumulator.append(structure)
        if len(accumulator) == 1:
            return accumulator[0]
        else:
            return accumulator


class UnicarbKBRDFClient(RDFClientBase):
    predicate_processor_map = ChainFunctionDict()
    _predicates_seen = set()
    _sparql_endpoint_uri = 'http://203.101.226.16:40935/unicarbkbv2.0.1/query'

    def __init__(self):
        super(UnicarbKBRDFClient, self).__init__(self._sparql_endpoint_uri, NSUnicarbKBStructure)
        self.bind("glytoucan", NSGlyTouCan)
        self.bind("glycomedb", NSGlycomeDB)
        self.bind("glycan", NSGlycan)
        self.bind("glycoinfo", NSGlycoinfo)
        self.bind("skos", NSSKOS)
        self.bind("uniprot", NSUniprotCore)
        self.bind("gco", NSGlycoconjugate)
        self.bind('ukp', NSUnicarbKBProtein)
        self.bind("uk", NSUnicarbKB)
        self.bind("sio", NSSIO)
        self.bind("unicorn", NSUnicorn)
        self.bind("owl", NSOwl)
        self.bind("dcterms", NSDCTerms)
        self.bind("faldo", NSFaldo)


GlySpaceRDFClient = GlyTouCanRDFClient


glytoucan_client = client = GlyTouCanRDFClient()
unikarbkb_client = UnicarbKBRDFClient()


get = client.get
query = client.query
structure = client.structure
from_taxon = client.from_taxon
structures_with_motif = client.structures_with_motif
triples = client.triples


@GlyTouCanRDFClient.register_predicate_processor(NSGlycan.has_glycosequence)
@UnicarbKBRDFClient.register_predicate_processor(NSGlycan.has_glycosequence)
def has_glycosequence_processor(state, uri):
    """Detect and extract GlycoCT sequence data and parse
    into a |Glycan| object.

    Parameters
    ----------
    state : ReferenceEntity or dict
        The key-value store to add annotation to.
    uri : rdflib.term.URIRef
        The `URIRef` to load structure data from.

    Returns
    -------
    BoundURIRef
    """
    reference = uri(query_type='subject')
    # print(uri, reference, state)
    try:
        in_format = (reference.in_carbohydrate_format)
    except AttributeError:
        return uri
    if in_format == (NSGlycan.carbohydrate_format_glycoct):
        # trailing underscore in case a URI would claim "structure"
        state["structure_"] = [glycoct.loads(reference.has_sequence)]
    if "structure_" not in state and in_format == (NSGlycan.carbohydrate_format_wurcs):
        try:
            state["structure_"] = [wurcs.loads(reference.has_sequence)]
        except (wurcs.WURCSError, KeyError):
            pass
    if "structure_" not in state and in_format == (NSGlycan.carbohydrate_format_iupac_extended):
        try:
            state["structure_"] = [iupac.loads(reference.has_sequence)]
        except (iupac.IUPACError, KeyError):
            pass
    return uri


class ImageResource(dict):
    """ImageResource wraps image resource URIs provided by the data source,
    unifying the image fetching API.

    ImageResource inherits from |dict|, and its keys
    are pairs of (symbol_format, image_format), e.g. ("cgf", "png").

    The wrapping :meth:`get` method will fetch the relevant URI from
    storage and fetches the image data, which can be written to file.

    Attributes
    ----------
    _uriref_list: list
        List of wrapped resource URIs.
    """
    def __init__(self):
        dict.__init__(self)
        self._uriref_list = []

    def _materialize(self):
        for uriref in self._uriref_list:
            reference = uriref()
            image_format = reference.format.split("/")[1]
            symbol_format = split_uri(reference.has_symbol_format)[1].replace("symbol_format_", "")
            url = uriref.toPython()
            self[symbol_format, image_format] = url

    def register_uri(self, uriref):
        self._uriref_list.append(uriref)
        reference = uriref()
        image_format = reference.format.split("/")[1]
        symbol_format = split_uri(reference.has_symbol_format)[1].replace("symbol_format_", "")
        url = uriref.toPython()
        self[symbol_format, image_format] = url

    def get(self, symbol_format, image_format='png'):
        """Fetch image data from a remote source.

        Parameters
        ----------
        symbol_format : str
            The symbol format of the image to fetch.
        image_format : str, optional
            The image format to request. Defaults to 'png'. The supported
            formats will vary depending upon the remote source.

        Returns
        -------
        name : TYPE
            Description
        """
        url = self[symbol_format, image_format]
        return requests.get(url, verify=ssl_verification)

    def __repr__(self):
        return "ImageResource(%s)" % ', '.join(map(str, self.keys()))


# @GlyTouCanRDFClient.register_predicate_processor(NSGlycan.has_image)
def has_image_processor(state, uri):
    '''
    Intercept raw URIRefs to image generation services and store them in
    the :class:`ImageResource` instance associated with `state`. Since the
    predicate name is being used by the ImageResource, return `None` so
    the URIRef doesn't try to overwrite it.
    '''
    if "has_image" in state:
        image_resource_manager = state["has_image"][0]
    else:
        image_resource_manager = state["has_image"] = [ImageResource()]
        image_resource_manager = image_resource_manager[0]
    image_resource_manager.register_uri(uri)
    return None



from glypy.utils import invert_dict

# BaseTypeForRelativeConfiguration
# "3" and "4" in relative configuration are replace from
# "1" and "2" in "D" absolute configuration, respectively
base_type_to_descriptors = {
    "gro": "x",  # "xgro" has no relative configuration
    "thr": "34",
    "ery": "44",
    "ara": "344",
    "rib": "444",
    "lyx": "334",
    "xyl": "434",
    "all": "4444",
    "alt": "3444",
    "man": "3344",
    "glc": "4344",
    "gul": "4434",
    "ido": "3434",
    "tal": "3334",
    "gal": "4334"
}


descriptors_to_base_type = invert_dict(base_type_to_descriptors)



import json
from collections import namedtuple, defaultdict, deque
try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

from glypy.io import glycoct
from glypy.structure.glycan_composition import HashableGlycanComposition


EnzymeEdge = namedtuple("EnzymeEdge", ("parent", "child", "enzyme"))


def _enzyme_graph_inner():
    return defaultdict(set)


class EnzymeGraph(Mapping):
    def __init__(self, graph=None, seeds=None, metadata=None):
        if graph is None:
            graph = defaultdict(_enzyme_graph_inner)
        self.graph = graph
        self.seeds = set()
        if seeds is None:
            seeds = self.parentless()
        self.seeds.update(seeds)
        self.metadata = metadata or {}

    def __getitem__(self, key):
        return self.graph[key]

    def __setitem__(self, key, value):
        self.graph[key] = value

    def add(self, parent, child, enzyme):
        self[parent][child].add(enzyme)

    def __iter__(self):
        return iter(self.edges())

    def __len__(self):
        return self.edge_count()

    def clone(self):
        graph = defaultdict(_enzyme_graph_inner)
        for outer_key, outer_value in self.graph.items():
            for inner_key, inner_value in outer_value.items():
                graph[outer_key][inner_key] = inner_value.copy()
        return self.__class__(graph, self.seeds.copy())

    def nodes(self):
        acc = set()
        acc.update(self.graph)
        for i, v in enumerate(self.graph.values()):
            acc.update(v)
        return acc

    def edges(self):
        edges = set()
        for outer_key, outer_value in self.graph.items():
            for inner_key, inner_value in outer_value.items():
                for val in inner_value:
                    edges.add(EnzymeEdge(outer_key, inner_key, val))
        return edges

    def node_count(self):
        acc = set()
        acc.update(self.graph)
        for i, v in enumerate(self.graph.values()):
            acc.update(v)
        return len(acc)

    def edge_count(self):
        edges = 0
        for outer_key, outer_value in self.graph.items():
            for inner_key, inner_value in outer_value.items():
                edges += len(inner_value)
        return edges

    def __repr__(self):
        return "{}({:d})".format(self.__class__.__name__, self.node_count())

    def enzymes(self):
        enzyme_set = set()
        for outer_key, outer_value in self.graph.items():
            for inner_key, inner_value in outer_value.items():
                enzyme_set.update(inner_value)
        return enzyme_set

    def remove_enzyme(self, enzyme):
        edges_removed = list()
        for outer_key, outer_value in list(self.graph.items()):
            for inner_key, inner_value in list(outer_value.items()):
                if enzyme in inner_value:
                    inner_value.remove(enzyme)
                    edges_removed.append((outer_key, inner_key))
                if not inner_value:
                    outer_value.pop(inner_key)
                if not outer_value:
                    self.graph.pop(outer_key)
        nodes_to_remove = self.parentless() - self.seeds
        while nodes_to_remove:
            for node in nodes_to_remove:
                self.remove(node)
            nodes_to_remove = self.parentless() - self.seeds
        return edges_removed

    def parents(self, target):
        parents = []
        for outer_key, outer_value in self.graph.items():
            for inner_key, inner_value in outer_value.items():
                if inner_key == target:
                    parents.append(outer_key)
        return parents

    def parentless(self):
        is_parent = set(self.graph)
        is_parented = set()
        for i, v in enumerate(self.graph.values()):
            is_parented.update(v)
        return is_parent - is_parented

    def children(self, target):
        children = []
        children.extend(self.graph[target])
        return children

    def remove(self, prune):
        items = deque([prune])
        i = 0
        while items:
            node = items.popleft()
            if node in self.graph:
                i += 1
                self.graph.pop(node)
        return i

    def _dump_entity(self, entity):
        return str(entity)

    def _dump(self):
        data_structure = {
            "seeds": sorted([str(sd) for sd in self.seeds]),
            "enzymes": sorted(self.enzymes()),
            "graph": {},
            "metadata": {}
        }
        outgraph = {}
        for outer_key, outer_value in self.graph.items():
            outgraph_inner = dict()
            for inner_key, inner_value in outer_value.items():
                outgraph_inner[self._dump_entity(inner_key)] = list(inner_value)
            outgraph[self._dump_entity(outer_key)] = outgraph_inner
        data_structure['graph'] = outgraph
        return data_structure

    def dump(self, fh):
        d = self._dump()
        json.dump(d, fh, sort_keys=True, indent=2)

    def dumps(self):
        d = self._dump()
        return json.dumps(d, sort_keys=True, indent=2)

    @classmethod
    def _load_entity(self, entity):
        return entity

    @classmethod
    def _load(cls, data_structure):
        seeds = {cls._load_entity(sd) for sd in data_structure["seeds"]}
        graph = defaultdict(_enzyme_graph_inner)
        for outer_key, outer_value in data_structure["graph"].items():
            outgraph_inner = _enzyme_graph_inner()
            for inner_key, inner_value in outer_value.items():
                outgraph_inner[cls._load_entity(inner_key)] = set(inner_value)
            graph[cls._load_entity(outer_key)] = outgraph_inner
        metadata = data_structure.get('metadata')
        inst = cls(graph, seeds, metadata)
        return inst

    @classmethod
    def loads(cls, text):
        data = json.loads(text)
        return cls._load(data)

    @classmethod
    def load(cls, fd):
        data = json.load(fd)
        return cls._load(data)

    def __eq__(self, other):
        return self.graph == other.graph

    def __ne__(self, other):
        return self.graph != other.graph

    def items(self):
        return self.graph.items()

    def keys(self):
        return self.graph.keys()

    def values(self):
        return self.graph.values()

    def merge(self, other):
        for parent, children in other.items():
            for child, enzymes in children.items():
                self[parent][child].update(enzymes)

    def _dijkstra_distances_and_paths(self, source, sink):
        distances = dict()
        previous = dict()
        unvisited = set()
        for node in self.nodes():
            distances[node] = float('inf')
            previous[node] = None
            unvisited.add(node)

        distances[source] = 0
        unvisited_finite_distance = dict()
        visit_queue = deque([source])
        while sink in unvisited:
            try:
                current_node = visit_queue.popleft()
            except IndexError:
                if unvisited_finite_distance:
                    current_node, _ = min(unvisited_finite_distance.items(), key=lambda x: x[1])
                    unvisited_finite_distance.pop(current_node)
                else:
                    current_node, _ = min(distances.items(), key=lambda x: x[1])
            try:
                unvisited.remove(current_node)
            except KeyError:
                continue
            for child in self.children(current_node):
                # all edges are of length 1
                alternate_distance = distances[current_node] + 1
                if alternate_distance < distances[child]:
                    distances[child] = alternate_distance
                    previous[child] = (current_node, self[current_node][child])
                if child in unvisited:
                    unvisited_finite_distance[child] = alternate_distance

        return distances, previous

    def path_between(self, source, sink):
        _, previous = self._dijkstra_distances_and_paths(source, sink)
        parent, enz = previous[sink]
        path = []
        path.append(EnzymeEdge(parent, sink, enz))
        child = parent
        while source != child:
            parent, enz = previous[child]
            path.append(EnzymeEdge(parent, child, enz))
            child = parent
        return path[::-1]


# This may be too memory intensive to use on large graphs because
# a single :class:`~.Glycan` instance uses many times the memory that
# a :class:`~.GlycanComposition` does.
class GlycanStructureEnzymeGraph(EnzymeGraph):

    @classmethod
    def _load_entity(self, entity):
        return glycoct.loads(entity)


class GlycanCompositionEnzymeGraph(EnzymeGraph):

    @classmethod
    def _load_entity(self, entity):
        return HashableGlycanComposition.parse(entity)



import operator
from collections import defaultdict
import functools

from six import string_types as basestring

from glypy import Substituent, monosaccharides
from glypy.structure.constants import Modification, Stem, UnknownPosition


class NodeSimilarityComparator(object):
    '''A heuristic comparison for measuring similarity between monosaccharides.

    Compares:
        1. ring_start and ring_end
        2. superclass
        3. configuration
        4. stem
        5. anomer
        6. If `include_modifications`, each modification
        7. If `include_substituents`, each substituent
        8. If `include_children`, each child |Monosaccharide|

    Attributes
    ----------
    include_substituents: bool
        Include substituents in comparison (Defaults |True|)
    include_modifications: bool
        Include modifications in comparison (Defaults |True|)
    include_children: bool
        Include children in comparison (Defaults |False|)
    exact: bool
        Penalize for having unmatched attachments (Defaults |True|)
    ignore_reduction: bool
        Whether or not to include differences in reduction state as a
        mismatch
    ignore_ring: bool
        Whether or not to include differences in ring coordinates as
        a mismatch
    treat_null_as_wild: bool
        Whether or not to treat traits with a value of :const:`None` or
        :const:`~.UnknownPosition` as always matching when the null
        value is on the *target* residue (the residue that traits are being
        matched to).
    short_circuit_after: None or Number
        Controls whether to quit comparing nodes if the difference
        becomes too large, useful for speeding up pessimistic
        comparisons
    visited: set
        Tracks which node pairs have already been compared to break
        cycles. This carries state across multiple calls to :meth:`compare`
        and must be reset by calling :meth:`reset` before reusing an
        instance on new structures.
    '''
    def __init__(self, include_substituents=True, include_modifications=True,
                 include_children=False, exact=True, ignore_reduction=False,
                 ignore_ring=False, treat_null_as_wild=True,
                 match_attachement_positions=False, short_circuit_after=None,
                 visited=None):
        if visited is None:
            visited = set()
        self.include_substituents = include_substituents
        self.include_modifications = include_modifications
        self.include_children = include_children
        self.exact = exact
        self.ignore_ring = ignore_ring
        self.ignore_reduction = ignore_reduction
        self.treat_null_as_wild = treat_null_as_wild
        self.match_attachement_positions = match_attachement_positions
        self.visited = visited
        self.short_circuit_after = short_circuit_after

    def reset(self):
        self.visited.clear()

    @classmethod
    def similarity(cls, node, target, include_substituents=True,
                   include_modifications=True, include_children=False,
                   exact=True, ignore_reduction=False, ignore_ring=False,
                   treat_null_as_wild=True, match_attachement_positions=False,
                   short_circuit_after=None, visited=None):
        """A heuristic comparison for measuring similarity between monosaccharides.

        Compares:
            1. ring_start and ring_end
            2. superclass
            3. configuration
            4. stem
            5. anomer
            6. If `include_modifications`, each modification
            7. If `include_substituents`, each substituent
            8. If `include_children`, each child |Monosaccharide|

        The result is two numbers, the observed similarity between `node` and `target`,
        and the similarity between `target` and itself. `expected - observed` is there
        number of differences observed between the two monosaccharides, which can be useful
        for expressing how far apart two monosaccharides are in feature space. For more distant
        similarity testing, especially when considering children, the ratio `observed / expected`
        might be used instead.

        Similarity is not symmetric, e.g. ``a -> b != b -> a``. A commutative version of similarity
        can be used by calculating both directions, and taking the result with the smallest error.

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against
        include_substituents: bool
            Include substituents in comparison (Defaults |True|)
        include_modifications: bool
            Include modifications in comparison (Defaults |True|)
        include_children: bool
            Include children in comparison (Defaults |False|)
        exact: bool
            Penalize for having unmatched attachments (Defaults |True|)
        ignore_reduction: bool
            Whether or not to include differences in reduction state as a
            mismatch
        ignore_ring: bool
            Whether or not to include differences in ring coordinates as
            a mismatch
        treat_null_as_wild: bool
            Whether or not to treat traits with a value of :const:`None` or
            :const:`~.UnknownPosition` as always matching when the null
            value is on the *target* residue (the residue that traits are being
            matched to).
        short_circuit_after: None or Number
            Controls whether to quit comparing nodes if the difference
            becomes too large, useful for speeding up pessimistic
            comparisons
        visited: set
            Tracks which node pairs have already been compared to break
            cycles. This carries state across multiple calls to :meth:`compare`
            and must be reset by calling :meth:`reset` before reusing an
            instance on new structures.

        Returns
        -------
        :class:`int`: observed
            The number of observed features that matched
        :class:`int`: expected
            The number of features that could have been matched
        """
        inst = cls(
            include_substituents=include_substituents,
            include_modifications=include_modifications,
            include_children=include_children,
            exact=exact, ignore_reduction=ignore_reduction,
            ignore_ring=ignore_ring, treat_null_as_wild=treat_null_as_wild,
            match_attachement_positions=match_attachement_positions,
            short_circuit_after=short_circuit_after,
            visited=visited)
        return inst.compare(node, target)

    def compare_anomer(self, node, target):
        """Compare :attr:`~.Monosaccharide.anomer` of `node` and `target`

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        test: int
            The similarity between node and target
        reference: int
            The expected similarity between target and itself
        """
        test = (node.anomer == target.anomer) or ((target.anomer.value is None) and self.treat_null_as_wild)
        reference = 1
        return int(test), reference

    def compare_compositions(self, node, target):
        """Compre :meth:`~.Monosaccharide.total_composition` of `node` and `target`.

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        test: int
            The similarity between node and target
        reference: int
            The expected similarity between target and itself
        """
        test = int(node.total_composition() == target.total_composition())
        reference = 1
        return test, reference

    def compare_ring_structure(self, node, target):
        """Compare the :attr:`~.Monosaccharide.superclass`, :attr:`~.Monosaccharide.stem`,
        :attr:`~.Monosaccharide.configuration`, and possibly :attr:`~.Monosaccharide.ring_start`
        and :attr:`~.Monosaccharide.ring_end` of `node` and `target`.

        If :attr:`ignore_ring` is :const:`True`, ring positions will be assumed to match.

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        test: int
            The similarity between node and target
        reference: int
            The expected similarity between target and itself
        """
        test = reference = 0
        test += (node.superclass == target.superclass) or ((target.superclass.value is None) and
                                                           self.treat_null_as_wild)
        reference += 1
        test += (node.stem == target.stem) or ((target.stem[0].value is None) and self.treat_null_as_wild)
        reference += 1
        test += (node.configuration == target.configuration) or ((target.configuration[0].value is None) and
                                                                 self.treat_null_as_wild)
        reference += 1
        if not self.ignore_ring:
            test += (node.ring_start == target.ring_start) or ((target.ring_start == UnknownPosition) and
                                                               self.treat_null_as_wild)
            reference += 1
            test += (node.ring_end == target.ring_end) or ((target.ring_end == UnknownPosition) and
                                                           self.treat_null_as_wild)
            reference += 1
        else:
            test += 2
            reference += 2
        return test, reference

    def compare_modifications(self, node, target):
        """Compare the modifications of `node` and `target`

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        test: int
            The similarity between node and target
        reference: int
            The expected similarity between target and itself
        """
        test = reference = 0
        node_reduced = False
        target_reduced = False
        n_mods = 0
        if self.match_attachement_positions:
            node_mods = node.modifications
            n_mods = len(node_mods)
            for pos, mod in target.modifications.items():
                if mod == 'aldi':
                    target_reduced = True
                check = (mod in node_mods[pos])
                if check:
                    if mod == 'aldi':
                        node_reduced = True
                    test += 1
                    n_mods -= 1
                reference += 1
        else:
            node_mods = list(node.modifications.values())
            n_mods = len(node_mods)
            for mod in target.modifications.values():
                if mod == 'aldi':
                    target_reduced = True
                check = (mod in node_mods)
                if check:
                    if mod == 'aldi':
                        node_reduced = True
                    test += 1
                    node_mods.pop(node_mods.index(mod))
                    n_mods -= 1
                reference += 1

        if self.ignore_reduction:
            if target_reduced:
                reference -= 1
            if node_reduced:
                test -= 1
        reference += n_mods if self.exact else 0
        return test, reference

    def compare_substituents(self, node, target):
        """Compare the substituents of `node` and `target`.

        If :attr:`match_attachement_positions` is :const:`True`,
        the positions must match exactly, otherwise only the substituent
        types must match.

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        test: int
            The similarity between node and target
        reference: int
            The expected similarity between target and itself
        """
        test = reference = 0
        if self.match_attachement_positions:
            node_subs = defaultdict(list)
            n_subs = 0
            for p, sub in node.substituents():
                n_subs += 1
                node_subs[p].append(sub)
            for pos, sub in target.substituents():
                if (sub in node_subs[pos]):
                    test += 1
                    n_subs -= 1
                reference += 1
        else:
            node_subs = [node_sub for p, node_sub in node.substituents()]
            n_subs = len(node_subs)
            for pos, sub in target.substituents():
                if (sub in node_subs):
                    test += 1
                    node_subs.pop(node_subs.index(sub))
                    n_subs -= 1
                reference += 1
        reference += n_subs if self.exact else 0
        return test, reference

    def _build_child_pair_score_map(self, node, target):
        '''Compute all pair-wise similarities between children
        of ``node`` and ``target``

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        dict
        '''
        # TODO: support exactness penalty here, maybe recursively?
        node_children = list(child for p, child in node.children())
        match_index = dict()
        for p, target_child in target.children():
            for node_child in node_children:
                c_res, c_qs = self.compare(node_child, target_child)
                match_index[node_child.id, target_child.id] = (c_res, c_qs)
        return match_index

    def compare_children(self, node, target):
        """Compute the similarity of the set of children between `node` and `target`

        If :attr:`match_attachement_positions` is :const:`True`, this will require
        the positions of child nodes to match exactly, otherwise, all pair-wise combinations
        will be considered and the optimal solution will be selected using :meth:`optimal_assignment`

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        test: int
            The similarity between node and target
        reference: int
            The expected similarity between target and itself
        """
        test = reference = 0
        if self.match_attachement_positions:
            node_children = defaultdict(list)
            for pos, child in node.children():
                node_children[pos].append(child)
            target_children = defaultdict(list)
            for pos, child in target.children():
                target_children[pos].append(child)
            for pos, children in target_children.items():
                for t_child, n_child in zip(children, node_children[pos]):
                    test_child, reference_child = self.compare(t_child, n_child)
                    test += test_child
                    reference += reference_child
        else:
            match_index = self._build_child_pair_score_map(node, target)
            assignments = self.optimal_assignment(match_index)
            for ix in assignments:
                a_test, a_reference = match_index[ix]
                test += a_test
                reference += a_reference
        return test, reference

    def _check_short_circuit(self, test, reference):
        return self.short_circuit_after is not None and (test - reference) < self.short_circuit_after

    def compare(self, node, target):
        """Calculate the similarity between `node` and `target`.

        This method does most of the organizational work, calling the appropriate
        methods and checking for short-circuiting.

        Parameters
        ----------
        node : :class:`~.Monosaccharide`
            The reference monosaccharide
        target : :class:`~.Monosaccharide`
            The monosaccharide to compare against

        Returns
        -------
        test: int
            The similarity between node and target
        reference: int
            The expected similarity between target and itself
        """
        key = (node.id, target.id)
        if key in self.visited:
            return 0, 0
        self.visited.add(key)
        test = 0
        reference = 0
        try:
            t, r = self.compare_anomer(node, target)
            test += t
            reference += r
        except AttributeError:
            # must be handling substituents
            t, r = self.compare_compositions(node, target)
            test += t
            reference += r
            return test, reference
        t, r = self.compare_ring_structure(node, target)
        test += t
        reference += r
        if self._check_short_circuit(test, reference):
            return test, reference
        if self.include_modifications:
            t, r = self.compare_modifications(node, target)
            test += t
            reference += r
            if self._check_short_circuit(test, reference):
                return test, reference
        if self.include_substituents:
            t, r = self.compare_substituents(node, target)
            test += t
            reference += r
            if self._check_short_circuit(test, reference):
                return test, reference
        if self.include_children:
            t, r = self.compare_children(node, target)
            test += t
            reference += r
        return test, reference

    def optimal_assignment(self, assignments):
        '''
        Given a set of possibly overlapping matches, find the
        optimal solution.
        '''
        diff_map = dict()
        for ids in assignments:
            diff_map[ids] = operator.sub(*assignments[ids])

        index_pair_sets = self.build_unique_index_pairs(assignments)

        def ordering_fn(index_pairs):
            total = 0
            for ix in index_pairs:
                total += max(assignments[ix])
            return total

        ordered_index_pairs_sets = sorted(index_pair_sets, key=ordering_fn, reverse=True)

        best_score = -float('inf')
        best_mapping = {}
        # prefer solutions which have a higher maximum value (more points of comparison)
        for assignment in ordered_index_pairs_sets:
            current_score = 0
            for ix in assignment:
                current_score += diff_map[ix]
            if current_score > best_score:
                best_score = current_score
                best_mapping = assignment
        return best_mapping

    def build_unique_index_pairs(self, pairs):
        '''
        Generate all unique non-overlapping sets of pairs, given in
        `pairs`
        '''
        depth = 0
        pairings = defaultdict(set)
        for a, b in pairs:
            pairings[a].add(b)
        next_current = [()]
        options_a = list(pairings)
        partial_solutions = set()
        while depth < len(options_a):
            for current in next_current:
                if len(current) > 0:
                    current_a, current_b = map(set, zip(*current))
                else:
                    current_b = set()
                components = set()
                a = options_a[depth]
                for b in pairings[a] - current_b:
                    components.add((current + ((a, b),)))
                partial_solutions.update(components)
            depth += 1
            next_current = partial_solutions
            partial_solutions = set()
        return list(next_current)


monosaccharide_similarity = NodeSimilarityComparator.similarity


def commutative_similarity(node, target, tolerance=0, *args, **kwargs):
    """Apply :func:`monosaccharide_similarity` to ``node`` and ``target`` for both
    ``node --> target`` and ``target --> node``, returning whether either comparison
    passes the tolerance threshold.

    Parameters
    ----------
    node: :class:`~.Monosaccharide`
        The reference monosaccharide
    target: :class:`~.Monosaccharide`
        The monosaccharide to compare against
    tolerance: :class:`int`, optional
        The minimum number of errors to tolerate
    *args:
        Forwarded to :func:`monosaccharide_similarity`
    **kwargs:
        Forwarded to :func:`monosaccharide_similarity`

    Returns
    -------
    :class:`bool`
    """
    obs, expect = monosaccharide_similarity(node, target, *args, **kwargs)
    if (expect - obs) <= tolerance:
        return True
    else:
        obs, expect = monosaccharide_similarity(target, node, *args, **kwargs)
        return (expect - obs) <= tolerance


def commutative_similarity_score(node, target, *args, **kwargs):
    """Apply :func:`monosaccharide_similarity` to ``node`` and ``target`` for both
    ``node --> target`` and ``target --> node``, returning the maximally normalized
    ratio of `observed / expected`.

    Parameters
    ----------
    node: :class:`~.Monosaccharide`
        The reference monosaccharide
    target: :class:`~.Monosaccharide`
        The monosaccharide to compare against
    *args:
        Forwarded to :func:`monosaccharide_similarity`
    **kwargs:
        Forwarded to :func:`monosaccharide_similarity`

    Returns
    -------
    :class:`float`:
        The maximal similarity score ratio
    """
    a_b, b_b = monosaccharide_similarity(node, target, *args, **kwargs)
    b_a, a_a = monosaccharide_similarity(target, node, *args, **kwargs)
    return max(a_b / (1. * b_b), b_a / (1. * a_a))


def commutative_similarity_score_with_tolerance(node, target, tolerance, *args, **kwargs):
    """Apply :func:`monosaccharide_similarity` to ``node`` and ``target`` for both
    ``node --> target`` and ``target --> node``, returning the maximally normalized
    ratio score, and whether there was a pair error less than `tolerance`.

    This can be viewed as a combination of :func:`commutative_similarity` and
    :func:`commutative_similarity_score` while making fewer calls to
    :func:`monosaccharide_similarity`.

    Parameters
    ----------
    node: :class:`~.Monosaccharide`
        The reference monosaccharide
    target: :class:`~.Monosaccharide`
        The monosaccharide to compare against
    tolerance: :class:`int`
        The minimum number of errors to tolerate
    *args:
        Forwarded to :func:`monosaccharide_similarity`
    **kwargs:
        Forwarded to :func:`monosaccharide_similarity`

    Returns
    -------
    :class:`float`:
        The maximal similarity score ratio
    :class:`bool`:
        Whether the difference passes error tolerance
    """
    a_b, b_b = monosaccharide_similarity(node, target, *args, **kwargs)
    b_a, a_a = monosaccharide_similarity(target, node, *args, **kwargs)
    pass_threshold = False
    if (b_b - a_b) <= tolerance or (a_a - b_a) <= tolerance:
        pass_threshold = True
    return max(a_b / (1. * b_b), b_a / (1. * a_a)), pass_threshold



def has_substituent(monosaccharide, substituent):
    """Checks whether ``monosaccharide`` has any substituent groups
    matching ``substituent``.

    Parameters
    ----------
    monosaccharide : :class:`~.Monosaccharide`
        The monosaccharide to check
    substituent : :class:`~.Substituent` or :class:`str`
        The substituent to check for

    Returns
    -------
    :class:`bool`
    """
    # Use the setter property to force the translation
    # of the name string.
    if isinstance(substituent, basestring):
        substituent = Substituent(substituent)
    substituent = substituent._name
    for position, subst in monosaccharide.substituents():
        if substituent == subst._name:
            return True
    return False


def has_modification(monosaccharide, modification):
    """Checks whether ``monosaccharide`` has any modification sites
    matching ``modification``.

    Parameters
    ----------
    monosaccharide : :class:`~.Monosaccharide`
        The monosaccharide to check
    modification : :class:`~.constants.Modification` or :class:`str`
        The modification to check for

    Returns
    -------
    :class:`bool`
    """
    for position, mod in monosaccharide.modifications.items():
        if mod == modification:
            return True
    return False


def has_monosaccharide(glycan, monosaccharide, tolerance=0, *args, **kwargs):
    """Checks whether ``glycan`` has any monosaccharide nodes
    matching ``monosaccharide`` within ``tolerance`` using
    :func:`commutative_similarity`

    Parameters
    ----------
    glycan: :class:`~.SaccharideCollection`
        The glycan structure or composition to search
    monosaccharide: :class:`~.Monosaccharide`
        The monosaccharide to search for
    tolerance : int, optional
        The error tolerance to use
    *args:
        Forwarded to :func:`monosaccharide_similarity`
    **kwargs:
        Forwarded to :func:`monosaccharide_similarity`

    Returns
    -------
    :class:`bool`
    """
    if isinstance(monosaccharide, basestring):
        monosaccharide = monosaccharides[monosaccharide]
    visited = set()
    for node in glycan:
        if commutative_similarity(
                node, monosaccharide, tolerance=tolerance, visited=visited, *args, **kwargs):
            return node
    return False


def is_reduced(obj):
    """A simple predicate to test whether an object has a reduced structure.

     If `obj` does not have a `reducing_end` attribute, this will return :const:`False`

    Parameters
    ----------
    obj : object
        The object to check

    Returns
    -------
    bool
    """
    try:
        return obj.reducing_end is not None
    except AttributeError:
        return False


def is_amine(substituent):
    """A simple predicate to test whether a substituent has an amine group adjacent
    attached to the carbon backbone by naming convention.

    This predicate checks to see if the name of the substituent is "amino" or if it
    starts with the phrase "n_" only.

    Parameters
    ----------
    substituent : Substituent  or str
        The object to test

    Returns
    -------
    bool
    """
    if isinstance(substituent, Substituent):
        name = substituent.name
    else:
        name = substituent
    return name.startswith("n_") or name == "amino"


def is_aminated(monosaccharide):
    """Tests to see if any substituents of `monosaccharide` are amines.

    Each substituent is tested using :func:`is_amine`, with all the caveats that
    entails.

    Parameters
    ----------
    monosaccharide : Monosaccharide
        The monosaccharide to test

    Returns
    -------
    bool

    See Also
    --------
    is_amine
    """
    for p, substituent in monosaccharide.substituents():
        if is_amine(substituent):
            return True
    return False


has_fucose = functools.partial(has_monosaccharide, monosaccharide=monosaccharides["Fucose"])
has_n_acetyl = functools.partial(has_substituent, substituent=Substituent("n-acetyl"))
is_acidic = functools.partial(has_modification, modification=Modification.Acidic)
is_sulfated = functools.partial(has_substituent, substituent=Substituent("sulfate"))


def is_generic_monosaccharide(monosaccharide):
    """Tests if the :attr:`~.Monosaccharide.stem` is unknown.

    Parameters
    ----------
    monosaccharide : Monosaccharide
        The object to test

    Returns
    -------
    bool
    """
    return monosaccharide.stem[0] is Stem.x


def is_derivatized(monosaccharide):
    """Tests whether any of the substituents attached to `monosaccharide` were
    added by derivatization.

    Parameters
    ----------
    monosaccharide : Monosaccharide
        The object to test

    Returns
    -------
    bool
    """
    for pos, sub in monosaccharide.substituents():
        if sub._derivatize:
            return True
    return False



# pragma: no cover

import requests
from lxml import etree

template_url = "http://www.monosaccharidedb.org/display_monosaccharide.action?output=xml&id={0}"


def get(monosaccharide_id):
    r = requests.request(url=template_url.format(monosaccharide_id),
                         method='GET', headers={'Accept-Language': 'en-US,en;q=0.9'})
    r.raise_for_status()
    root = etree.fromstring(r.text)
    return _unpack_element(root)


def _local_name(element):
    return element.tag


schema_info = {
    'lists': {
        'atom',
        'connection',
        'position',
        'alias',
        'core_modification',
    },
    'strings': {
        'stereocode',
    }
}


def typify(val):
    if val in ("true", "false"):
        if val == 'true':
            return True
        elif val == 'false':
            return False
    try:
        return int(val)
    except Exception:
        try:
            return float(val)
        except Exception:
            return str(val)


def _unpack_element(element, **kwargs):
    try:
        name = kwargs.pop('ename')
    except KeyError:
        name = _local_name(element)
    info = {k: typify(v) for k, v in dict(element.attrib).items()}
    # process subelements
    for child in element.iterchildren():
        cname = _local_name(child)
        if cname not in schema_info['lists']:
            info[cname] = _unpack_element(child, ename=cname, **kwargs)
        else:
            info.setdefault(cname, []).append(_unpack_element(child, ename=cname, **kwargs))

    # process element text
    if element.text:
        stext = element.text.strip()
        if stext:
            val = typify(stext) if name not in schema_info['strings'] else stext
            if info:
                info[name] = val
            else:
                return val
    return info



"""Implementation of explicit sub-graph inclusion and co-traversal algorithms.

Graph inclusion here means that there exists a path through graph A that can be
mapped onto equivalent nodes in graph B that are connected in the same way.

"""
from collections import deque, defaultdict

from glypy.structure import UnknownPosition
from glypy.algorithms.similarity import commutative_similarity, commutative_similarity_score_with_tolerance
from glypy.utils import root


class TopologicalInclusionMatcher(object):
    """A recursive topological traversal.

    Attributes
    ----------
    target: :class:`~.Monosaccharide`
        The target to test for inclusion with
    reference: :class:`~.Monosaccharide`
        The reference to test for inclusion in
    substituents: bool
        Whether to examine substituents when computing similarity and inclusion.
    tolerance: float
        The magnitude of the similarity error to tolerate.
    visited: set
        The set of target and reference node id pairs that have already been considered.

    """
    def __init__(self, target, reference, substituents=True, tolerance=0, visited=None):
        self.target = target
        self.reference = reference
        self.substituents = substituents
        self.tolerance = tolerance
        self.visited = visited or set()

    @classmethod
    def compare(cls, target, reference, substituents=True, tolerance=0, visited=None):
        '''
        A generalization of :meth:`~Monosaccharide.topological_equality` which allows for ``target``
        to be matched to ``reference``, but for ``reference`` to include more. Consequently,
        this method is not commutative.

        Parameters
        ----------
        target: :class:`~.Monosaccharide`
            The monosaccharide to compare
        reference: :class:`~.Monosaccharide`
            The monosaccharide to compare against
        substituents: :class:`bool`, optional
            Whether or not to compare with the substituents of each node. Defaults
            to :const:`True`.
        tolerance: :class:`float`, optional
            The maximum difference to permit between nodes for inclusion. Defaults
            to 0

        Returns
        -------
        :class:`float`
            The inclusion score. Greater than 0 indicates topological inclusion,
            though larger corresponds to better alignment.

        See Also
        --------
        :func:`~.commutative_similarity_score_with_tolerance`
        exact_ordering_inclusion
        '''
        inst = cls(
            target, reference, substituents=substituents,
            tolerance=tolerance, visited=visited)
        score = inst.test()
        return score

    def test_similarity(self):
        """Calculate the commutative similarity between :attr:`target` and
        :attr:`reference`.

        Returns
        -------
        float

        See Also
        --------
        :func:`~.commutative_similarity_score_with_tolerance`
        """
        score, similar = commutative_similarity_score_with_tolerance(
            self.target, self.reference, self.tolerance,
            include_substituents=self.substituents)
        return score, similar

    def test_children(self):
        """Find the optimal similarity pairing between descendents of
        :attr:`target` and :attr:`reference`.

        """
        match_index = self._build_child_pair_score_map()
        required_nodes = {node.id for p, node in self.target.children()}
        optimal_assignment, score = self.optimal_assignment(match_index, required_nodes)
        return optimal_assignment, score, bool(required_nodes)

    def test(self):
        """Compute the similarity metric for the current node pair.

        Returns
        -------
        float:
            The similarity score for the current pair.
        """
        key = (self.target.id, self.reference.id)
        if key in self.visited:
            return True
        self.visited.add(key)
        pair_score, similar = self.test_similarity()
        if similar:
            child_pairs, score, had_children = self.test_children()
            if child_pairs:
                return pair_score + score
            elif not had_children:
                return pair_score
            else:
                return 0
        return 0

    def _build_child_pair_score_map(self):
        combinations = dict()
        for a_pos, a_child in self.target.children():
            for b_pos, b_child in self.reference.children():
                inclusion_score = self.compare(
                    a_child, b_child, substituents=self.substituents,
                    tolerance=self.tolerance, visited=self.visited)
                if inclusion_score > 0:
                    combinations[a_child.id, b_child.id] = inclusion_score
        return combinations

    def build_unique_index_pairs(self, pairs):
        '''
        Generate all unique non-overlapping sets of pairs, given in
        `pairs`
        '''
        depth = 0
        pairings = defaultdict(set)
        for a, b in pairs:
            pairings[a].add(b)
        next_current = [()]
        options_a = list(pairings)
        partial_solutions = set()
        while depth < len(options_a):
            for current in next_current:
                if len(current) > 0:
                    current_a, current_b = map(set, zip(*current))
                else:
                    current_b = set()
                components = set()
                a = options_a[depth]
                for b in pairings[a] - current_b:
                    components.add((current + ((a, b),)))
                partial_solutions.update(components)
            depth += 1
            next_current = partial_solutions
            partial_solutions = set()
        return list(next_current)

    def optimal_assignment(self, assignments, required_nodes=None):
        index_pair_sets = self.build_unique_index_pairs(assignments)
        best_score = -float('inf')
        best_mapping = None
        for assignment in index_pair_sets:
            current_score = 0
            has_nodes = set()
            for ix in assignment:
                current_score += assignments[ix]
                has_nodes.add(ix[0])
            if current_score > best_score and has_nodes == required_nodes:
                best_score = current_score
                best_mapping = assignment
        return best_mapping, best_score


topological_inclusion = TopologicalInclusionMatcher.compare


def exact_ordering_inclusion(target, reference, substituents=True, tolerance=0, visited=None):
    '''
    A generalization of :meth:`~Monosaccharide.exact_ordering_equality` which allows for ``target``
    to be matched to ``reference``, but for ``reference`` to include more. Consequently,
    this method is not commutative.

    Parameters
    ----------
    target: :class:`~.Monosaccharide`
        The monosaccharide to compare
    reference: :class:`~.Monosaccharide`
        The monosaccharide to compare against
    substituents: :class:`bool`, optional
        Whether or not to compare with the substituents of each node. Defaults
        to :const:`True`.
    tolerance: :class:`float`, optional
        The maximum difference to permit between nodes for inclusion. Defaults
        to 0

    Returns
    -------
    :class:`float`:
        The similarity score between the two structures. A score of 0 means no
        inclusion, and a non-zero score means inclusion, with larger scores indicating
        more node pairs matching.

    See Also
    --------
    commutative_similarity_score_with_tolerance
    topological_inclusion
    '''
    if visited is None:
        visited = set()
    if (target.id, reference.id) in visited:
        return True
    node_score, similar = commutative_similarity_score_with_tolerance(
        target, reference, tolerance, include_substituents=substituents)
    if similar:
        if substituents:
            reference_substituents = dict(reference.substituents())
            for a_pos, a_sub in target.substituents():
                b_sub = reference_substituents.get(a_pos)
                if b_sub is None:  # pragma: no cover
                    return 0
                if a_sub != b_sub:  # pragma: no cover
                    return 0
        reference_mods = dict(reference.modifications.items())
        for a_pos, a_mod in target.modifications.items():
            b_mod = reference_mods.get(a_pos)
            if b_mod is None:  # pragma: no cover
                return 0
            if a_mod != b_mod:  # pragma: no cover
                return 0
        reference_children = dict(reference.children())
        for pos, a_child in target.children():
            b_child = reference_children.get(pos)
            if b_child is None:  # pragma: no cover
                return 0
            if a_child[0] == b_child[0]:
                match_score = exact_ordering_inclusion(a_child, b_child, substituents=substituents,
                                                       tolerance=tolerance, visited=visited)
                if not match_score:
                    return 0
                node_score += match_score
            else:
                return 0
        return node_score


def subtree_of(subtree, tree, exact=False, include_substituents=True, tolerance=0):
    '''
    Test to see if `subtree` is included in `tree` anywhere. Returns the
    node id number of the first occurence of `subtree` included in `tree` or |None|
    if it is not found.

    Parameters
    ----------
    subtree: :class:`~.Glycan`
        The structure to search for. The search attempts to match the complete structure of subtree.
    tree: :class:`~.Glycan`
        The sturcture to search in. The search iterates over each residue in `tree` and calls a comparator
        function, comparing the `subtree` to the substructure rooted at that residue.
    exact: :class:`bool`
        If |True|, use :func:`exact_ordering_inclusion` to compare nodes. Otherwise use :func:`topological_inclusion`.
        Defaults to |False|.

    Returns
    -------
    |int| or |None| if no match
    '''
    if exact:
        comparator = exact_ordering_inclusion
    else:
        comparator = topological_inclusion
    tree_root = root(subtree)
    for node in tree:
        if comparator(tree_root, node, substituents=include_substituents, tolerance=tolerance):
            return node.id
    return None


def find_matching_subtree_roots(subtree, tree, exact=False, include_substituents=True, tolerance=0):
    '''
    Find the list of nodes where occurences of `subtree` included in `tree` are rooted.

    Parameters
    ----------
    subtree: :class:`~.Glycan`
        The structure to search for. The search attempts to match the complete structure of subtree.
    tree: :class:`~.Glycan`
        The sturcture to search in. The search iterates over each residue in `tree` and calls a comparator
        function, comparing the `subtree` to the substructure rooted at that residue.
    exact: :class:`bool`
        If |True|, use :func:`exact_ordering_inclusion` to compare nodes. Otherwise use :func:`topological_inclusion`.
        Defaults to |False|.

    Returns
    -------
    :class:`list` of :class:`~.Monosaccharide`
    '''
    if exact:
        comparator = exact_ordering_inclusion
    else:
        comparator = topological_inclusion
    tree_root = root(subtree)
    matched_nodes = []
    for node in tree:
        if comparator(tree_root, node, substituents=include_substituents, tolerance=tolerance):
            matched_nodes.append(node)
    return matched_nodes


def walk_with(query, reference, visited=None, comparator=commutative_similarity, include_substituents=True):
    """Walk the `query` along `reference`, yielding successive matched nodes along
    a subtree of `reference` using a :term:`Exact Matching` traversal.

    This function provides slightly more detail than :func:`subtree_of`
    as it exposes every step along the subgraph traversal, rather than just the root.


    Parameters
    ----------
    query : :class:`Glycan`
        The query structure to search with
    reference : :class:`Glycan`
        The reference structure to search in
    visited : :class:`set`, optional
        The set of node id pairs to ignore
    comparator : :class:`Callable`, optional
        The :class:`Callable` object which can compare two :class:`~.Monosaccharide`

    Yields
    ------
    (:class:`~.Monosaccharide`, :class:`~.Monosaccharide`)
        Pairs of matched nodes along a path
    """
    if visited is None:
        visited = set()
    query_root = root(query)
    reference_root = root(reference)
    node_stack = deque()
    if comparator(query_root, reference_root, include_substituents=include_substituents):
        node_stack.append((reference_root, query_root))
    while len(node_stack) != 0:
        rnode, qnode = node_stack.pop()
        key = (rnode.id, qnode.id)
        if key in visited:
            continue
        visited.add(key)
        # yield the next step along the matched path
        yield rnode, qnode

        for p, rlink in rnode.links.items():
            rparent = rlink.parent
            qparent = None
            rchild = rlink.child
            qchild = None
            # query link position is known
            if p != UnknownPosition:
                # check the parent of the link for matches, potentially flowing
                # through a cycle if one is present or flowing down the tree if
                # not starting at the root
                for qlink in qnode.links[p]:
                    if comparator(qlink.parent, rparent):
                        qparent = qlink.parent
                        break
                if qparent is not None:
                    key = (rparent.id, qparent.id)
                    if key not in visited:
                        node_stack.append((rparent, qparent))
                # check the child of the link for matches, proceding down the tree
                for qlink in qnode.links[p]:
                    if comparator(qlink.child, rchild):
                        qchild = qlink.child
                        break
                if qchild is not None:
                    key = (rchild.id, qchild.id)
                    if key not in visited:
                        node_stack.append((rchild, qchild))
            else:
                for qlink in qnode.links.values():
                    if comparator(qlink.parent, rparent) and comparator(qlink.child, rchild):
                        if rnode is rparent:
                            if (rchild.id, qlink.child.id) in visited:
                                continue
                        elif rnode is rchild:
                            if (rparent.id, qlink.parent.id) in visited:
                                continue
                        qparent = qlink.parent
                        qchild = qlink.child
                        break
                if qparent is not None:
                    key = (rparent.id, qparent.id)
                    if key not in visited:
                        node_stack.append((rparent, qparent))
                    key = (rchild.id, qchild.id)
                    if key not in visited:
                        node_stack.append((rchild, qchild))



import logging
try:
    from itertools import chain, izip_longest
except ImportError:
    from itertools import chain, zip_longest as izip_longest
from collections import deque

from six import string_types as basestring

from glypy.io.format_constants_map import anomer_map, superclass_map
from glypy.utils import (
    invert_dict,
    identity as ident_op,
    cyclewarning,
    uid,
    make_struct)
from glypy.utils.multimap import OrderedMultiMap
from glypy.utils.enum import EnumValue
from glypy.composition import Composition, calculate_mass
from glypy.composition.structure_composition import monosaccharide_composition
from glypy.composition.structure_composition import modification_compositions

from .constants import (
    Anomer, Configuration, Stem, SuperClass, Modification, RingType, UnknownPosition,
    NoPosition)
from .substituent import Substituent
from .link import Link
from .base import SaccharideBase
from .stereochemistry import stereocode


anomer_map = invert_dict(anomer_map)
superclass_map = invert_dict(superclass_map)

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")

debug = True


def _get_standard_composition(monosaccharide):
    '''Used to get initial composition for a given monosaccharide
    |Superclass| and modifications.

    Used during initialization of a |Monosaccharide|.

    Parameters
    ----------
    monosaccharide: :class:`Monosaccharide`
        The |Monosaccharide| object to read attributes from

    Returns
    -------
    :class:`~glypy.composition.composition.Composition`:
        The baseline composition from `monosaccharide.superclass` + `monosaccharide.modifications`
    '''
    base = monosaccharide_composition[monosaccharide.superclass.name]
    modifications = list(monosaccharide.modifications.items())
    double_bond_count = 0
    keto_group = False
    for mod_pos, mod_val in modifications:
        # Don't set the reducing end here
        if isinstance(mod_val, ReducedEnd):
            monosaccharide.reducing_end = mod_val
            continue
        # In case the modification is not properly transformed elsewhere
        elif mod_val is Modification.aldi:  # pragma: no cover
            monosaccharide.reducing_end = True
            continue
        elif mod_val is Modification.keto and mod_pos != 0:
            keto_group = True
        if mod_val == Modification.en:
            double_bond_count += 1
            if double_bond_count % 2 == 0:
                continue
        # Using global constant modification
        try:
            base += modification_compositions[mod_val](mod_pos)
        # Using object-oriented modification
        except Exception:  # pragma: no cover
            base += mod_val.composition

    # The presence of a keto group implicitly converts the C1 carbonyl oxygen into
    # a hydroxyl and balances C1 with another hydrogen.
    if keto_group:
        base += Composition("H2")
    return base


def traverse(monosaccharide, visited=None, apply_fn=ident_op):
    '''
    A low-level depth-first traversal method for unwrapped residue
    graphs

    Parameters
    ----------
    monosaccharide: :class:`Monosaccharide`
        Residue to start traversing from
    visited: set or None
        The collection of node ids to ignore, having already visited them. If |None|, it defaults
        to the empty set.
    apply_fn: function
        Function to apply to each residue before yielding them

    Yields
    -------
    :class:`Monosaccharide`
    '''
    if visited is None:
        visited = set()
    yield apply_fn(monosaccharide)
    visited.add(monosaccharide.id)
    outnodes = sorted(list(monosaccharide.links.items()), key=lambda x: x[1][monosaccharide].degree(), reverse=True)
    for link_pos, link in outnodes:
        child = link[monosaccharide]
        if child.id in visited:
            continue
        for grandchild in traverse(child, visited=visited, apply_fn=apply_fn):
            yield grandchild


def _traverse_debug(monosaccharide, visited=None, apply_fn=ident_op):  # pragma: no cover
    '''
    A low-level depth-first traversal method for unwrapped residue
    graphs when the :attr:`id` attribute may be masking duplicate residues

    Parameters
    ----------
    monosaccharide: :class:`Monosaccharide`
        Residue to start traversing from
    visited: set or None
        The collection of node ids to ignore, having already visited them. If |None|, it defaults
        to the empty set.
    apply_fn: function
        Function to apply to each residue before yielding them

    Yields
    -------
    :class:`Monosaccharide`
    '''
    if visited is None:
        visited = set()
    yield apply_fn(monosaccharide)
    visited.add(id(monosaccharide))
    outnodes = sorted(list(monosaccharide.links.items()), key=lambda x: x[1][monosaccharide].degree(), reverse=True)
    for link_pos, link in outnodes:
        child = link[monosaccharide]
        if id(child) in visited:
            continue
        for grandchild in _traverse_debug(child, visited=visited, apply_fn=apply_fn):
            yield grandchild


def graph_clone(monosaccharide, visited=None):
    '''
    Low-level depth-first duplication method for unwrapped residue graphs
    which contain monosaccharides with substituent parent nodes.

    Parameters
    ----------
    residue: :class:`Monosaccharide`
        The root of the graph to clone
    visited: set or None
        The collection of node ids to ignore, having already visited them. If |None|, it defaults
        to the empty set.

    Returns
    -------
    :class:`Monosaccharide`:
        The root of a newly duplicated and identical residue graph
    '''
    llen = len
    index = {}
    visited = set() if visited is None else visited
    clone_root = monosaccharide.clone(prop_id=True)
    index[clone_root.id] = clone_root
    node_stack = deque([(clone_root, monosaccharide)])
    node_stack_append = node_stack.append
    node_stack_pop = node_stack.pop
    while llen(node_stack) > 0:
        clone, ref = node_stack_pop()
        if ref.id in visited:
            continue
        visited.add(ref.id)
        links = [link for pos, link in ref.links.items()]
        for link in links:
            terminal = link[ref]
            if terminal.id in visited or terminal.node_type != Monosaccharide.node_type:
                continue
            # Handle cycles where the same node is linked many times
            if terminal.id in index:
                clone_terminal = index[terminal.id]
                # clone_terminal.maybe_cyclic = True
                cyclewarning()
            else:
                index[terminal.id] = clone_terminal = terminal.clone(prop_id=True)
            if link.is_child(terminal):
                link.clone(clone, clone_terminal)
            else:
                link.clone(clone_terminal, clone)
            node_stack_append((clone_terminal, terminal))
            i = 0
            for p, subst in clone_terminal.substituents():
                j = 0
                for sp, child in subst.children():
                    if child.node_type == Monosaccharide.node_type:
                        orig = terminal.substituent_links[p][i].child.children()[j][1]
                        node_stack_append((child, orig))
                    j += 1
                i += 1
    return clone_root


def release(monosaccharide):
    '''Break all monosaccharide-monosaccharide links on `monosaccharide`, returning
    them as a list. Breaking is done with `refund=True`

    Parameters
    ----------
    monosaccharide : Monosaccharide
        |Monosaccharide| to break all links on

    Returns
    -------
    list of tuple(link, (link.parent, link.child))
    '''
    return [(link, link.try_break_link(refund=True)) for link in list(monosaccharide.links.values())]


def toggle(monosaccharide):
    '''A simple generator for declaratively masking and masking a residue's links. The
    first iteration masks all links. The second unmasks them. Calls :func:`release`

    Parameters
    ----------
    monosaccharide : Monosaccharide
        |Monosaccharide| to mask links on
    '''
    links = release(monosaccharide)
    yield links
    for link, termini in links:
        link.try_apply()
    yield False


def depth(monosaccharide, visited=None):
    '''
    Calculate  the distance from `monosaccharide` to its furthest grand-child node.
    '''
    if visited is None:
        visited = set()
    if monosaccharide.id in visited:  # pragma: no cover
        return 0
    visited.add(monosaccharide.id)
    depth_count = 1
    if(len(monosaccharide.links) > 1):
        depth_count += max(depth(ch, visited) for p, ch in monosaccharide.children())
    return depth_count


class Monosaccharide(SaccharideBase):
    '''
    Represents a single monosaccharide molecule, and its relationships with other
    molcules through |Link| objects. |Link| objects stored in :attr:`links` for connections to other
    |Monosaccharide| instances, building a |Glycan|
    structure as a graph of |Monosaccharide| objects. |Link|
    objects connecting the |Monosaccharide| instance to |Substituent|
    objects are stored in :attr:`substituent_links`.

    Both :attr:`links` and :attr:`substituent_links` are instances of
    |OrderedMultiMap| objects where the key is the index of the
    carbon atom in the carbohydrate backbone that hosts the bond.
    An index of `x` or `-1` represents an unknown location.


    .. warning::

        While |Monosaccharide| objects expose their :attr:`.modifications`, :attr:`.links`, and
        :attr:`.substituent_links` attributes as mutable, you should treat them as **read-only**.
        The methods for altering their contents, :meth:`add_substituent`, :meth:`add_monosaccharide`,
        :meth:`add_modification`, :meth:`drop_substituent`, :meth:`drop_monosaccharide`, and
        :meth:`drop_modification` are all responsible for handling these mutations for you. |Link| methods
        like :meth:`Link.apply` and :meth:`Link.break_link` are used internally.


    Attributes
    ----------
    anomer: :class:`Anomer`
        An entry of :class:`~.Anomer` that corresponds to the linkage type
        of the carbohydrate backbone. Is an entry of a class based on :class:`Enum`
    superclass: :class:`~.SuperClass`
        An entry of :class:`~.SuperClass` that corresponds to the number of
        carbons in the carbohydrate backbone of the monosaccharide. Controls the base composition of the
        instance and the number of positions open to be linked to or modified. Is an entry of a class
        based on :class:`~.Enum`
    configuration: :class:`~.Configuration` or {'d', 'l', 'x', 'missing', None}
        An entry of |Configuration| which corresponds to the optical
        stereomer state of the instance. Is an entry of a class based on :class:`Enum`. May possess
        more than one value.
    stem: :class:`~.Stem`
        Corresponds to the bond conformation of the carbohydrate backbone. Is an entry of a class based
        on :class:`~.Enum`. May possess more than one value.
    ring_start: |int|
        The index of the carbon of the carbohydrate backbone that starts a ring. A value of ``-1``, ``'x'``, or
        |None| corresponds to an unknown start. A value of ``0`` refers to a linear chain.
    ring_end:  |int|
        The index of the carbon of the carbohydrate backbone that ends a ring. A value of ``-1``, ``'x'``, or
        |None| corresponds to an unknown ends. A value of ``0`` refers to a linear chain.
    stereocode: :class:`~.Stereocode`
        The stereochemistry of all carbons of the monosaccharide's backbone ring/chain.
    reducing_end: :class:`ReducedEnd`
        The reducing end terminal group of the monosaccharide if the monosaccharide is uncyclized
    modifications: |OrderedMultiMap|
        The mapping of sites to |Modification| entries. Directly modifies the instance's :attr:`composition`
    links: |OrderedMultiMap|
        The mapping of sites to |Link| entries that refer to other |Monosaccharide| instances
    substituent_links: |OrderedMultiMap|
        The mapping of sites to |Link| entries that refer to
        |Substituent| instances.
    composition: |Composition|
        An instance of |Composition| corresponding to the elemental composition
        of ``self`` and its immediate modifications. If not provided, this will be inferred
        from field values.
    reduced: :class:`ReducedEnd`
        An instance of ReducedEnd, or the value |True|, represents a reduced sugar. May be inferred
        from `modifications` if "aldi" is present

    '''
    _serializers = {}

    __slots__ = (
        "id", "_anomer", "_configuration", "_stem", "_superclass",
        "ring_start", "ring_end", "links", "substituent_links",
        "modifications", "composition",
        "_reducing_end", "_degree",
        "_checked_for_reduction"
    )

    def __init__(self, anomer=None, configuration=None, stem=None,
                 superclass=None, ring_start=UnknownPosition, ring_end=UnknownPosition,
                 modifications=None, links=None, substituent_links=None,
                 composition=None, reduced=None, id=None, fast=False):
        """Create a new :class:`Monosaccharide` using the provided attributes, leaving
        any unspecified attributes in an explicit unknown state.

        If ``fast`` is :const:`True`, all parameter values are assumed to be :class:`~.EnumValue`
        instances from the correct classes, and reduction checking is done quickly. This option
        is :const:`False` by default on explicit calls but set to :const:`True` for certain internal
        cases where this precondition can be ensured.

        Parameters
        ----------
        anomer : :class:`str` or :class:`~.EnumValue`, optional
            Specify the anomeric carbon state. Translates through :class:`~.Anomer`
        configuration : :class:`str` or :class:`~.EnumValue` or :class:`tuple` optional
            Specify the stereoisomeric configuration state(s) of the carbon rings. Will
            be converted to a :class:`tuple` if not provided as one, and elements will be
            translated through :class:`~.consants.Configuration`
        stem : :class:`str` or :class:`~.EnumValue` or :class:`tuple`, optional
            Specify the base carbon ring stereochemical template. Will
            be converted to a :class:`tuple` if not provided as one, and elements will be
            translated through :class:`~.consants.Stem`
        superclass : :class:`str` or :class:`~.EnumValue`, optional
            Specify the carbon chain length. Translates through :class:`~.SuperClass`.
        ring_start : int, optional
            Specify the carbon ring start position. If missing, defaults to :const:`UnknownPosition`
        ring_end : int, optional
            Specify the carbon ring closing position. If missing, defaults to :const:`UnknownPosition`
        modifications : OrderedMultiMap, optional
            Specify the collection of :class:`~.constant.Modification` values located at different carbon
            positions.
        links : OrderedMultiMap, optional
            Specify the collection of :class:`~.Link` instances connecting this monosaccharide to other
            monosaccharides
        substituent_links : OrderedMultiMap, optional
            Specify the collection of :class:`~.Link` instances connecting this monosaccharide to its
            substituents
        composition : :class:`~.Composition`, optional
            The chemical composition of the carbon ring to use prior to applying linkage changes and
            modifications. Automatically derived from `superclass` if omitted.
        reduced : :class:`ReducedEnd` or :class:`bool`, optional
            Whether the monosaccharide is reduced, or if a :class:`ReducedEnd` is given, specifies the
            reduction type.
        id : :class:`int`, optional
            The value to set for the node's :attr:`id`. If not provided, one will be generated using
            :func:`~.glypy.utils.uid`.
        fast : bool, optional
            Whether or not to use the fast initialization branch or not. Should only be used if all
            parameters have been properly typed and valid.
        """

        if modifications is None:  # pragma: no cover
            modifications = OrderedMultiMap()
        if links is None:
            links = OrderedMultiMap()

        if id is None:
            id = uid()

        self.modifications = modifications
        self._reducing_end = None
        self._checked_for_reduction = False

        if fast:
            self._anomer = anomer
            self._configuration = tuple(configuration)
            self._stem = tuple(stem)
            self._superclass = superclass
            self._fast_reduce(reduced)
        else:
            self.anomer = anomer
            self.configuration = configuration
            self.stem = stem
            self.superclass = superclass
            self.reducing_end = reduced

        self.ring_start = ring_start
        self.ring_end = ring_end
        self.links = links
        self.substituent_links = OrderedMultiMap() if substituent_links\
            is None else substituent_links
        self.id = id
        if composition is None:
            composition = _get_standard_composition(self)
        self.composition = composition
        self._degree = len(self.links) + len(self.substituent_links)

    @property
    def anomer(self):
        return self._anomer

    @anomer.setter
    def anomer(self, value):
        self._anomer = Anomer[value]

    @property
    def configuration(self):
        return self._configuration

    @configuration.setter
    def configuration(self, value):
        if isinstance(value, (tuple, list)):
            self._configuration = tuple(Configuration[v] for v in value)
        else:
            self._configuration = (Configuration[value],)

    @property
    def stem(self):
        return self._stem

    @stem.setter
    def stem(self, value):
        if isinstance(value, (tuple, list)):
            self._stem = tuple(Stem[v] for v in value)
        else:
            self._stem = (Stem[value],)

    @property
    def superclass(self):
        return self._superclass

    @superclass.setter
    def superclass(self, value):
        self._superclass = SuperClass[value]

    @property
    def stereocode(self):
        return stereocode(self)

    def __root__(self):
        return self

    def clone(self, prop_id=False, fast=True, monosaccharide_type=None):
        '''
        Copies just this |Monosaccharide| and its |Substituent|s, creating a separate instance
        with the same data. All mutable data structures are duplicated and distinct from the original.

        Does not copy any :attr:`links` as this would cause recursive duplication of the entire |Glycan|
        graph.

        Parameters
        ----------
        prop_id: :class:`bool`
            Whether to copy :attr:`id` from ``self`` to the new instance
        fast: :class:`bool`
            Whether to use the fast-path initialization process in :meth:`Monosaccharide.__init__`
        monosaccharide_type: :class:`type`
            A subclass of :class:`Monosaccharide` to use

        Returns
        -------
        :class:`Monosaccharide`

        '''
        if monosaccharide_type is None:
            monosaccharide_type = Monosaccharide
        modifications = OrderedMultiMap()
        for site, modification_list in self.modifications.lists():
            for mod in modification_list:
                if isinstance(mod, ReducedEnd):
                    continue
                try:
                    modifications[site] = mod
                except Exception:  # pragma: no cover
                    modifications[site] = mod.clone()

        monosaccharide = monosaccharide_type(
            superclass=self._superclass,
            stem=self._stem,
            configuration=self._configuration,
            ring_start=self.ring_start,
            ring_end=self.ring_end,
            modifications=modifications,
            anomer=self._anomer,
            reduced=self._reducing_end.clone(prop_id=prop_id) if self._reducing_end is not None else None,
            id=self.id if prop_id else None,
            fast=fast)
        for pos, link in self.substituent_links.items():
            sub = link.to(self)
            dup = sub.clone(prop_id=prop_id)
            link.clone(monosaccharide, dup)
        return monosaccharide

    @property
    def ring_type(self):
        """The size of the ring-shape of the carbohydrate, as computed
        by :attr:`ring_end` - :attr:`ring_start`.

        Returns
        -------
        :class:`~.EnumValue`:
            The appropriate value of :class:`~.RingType`
        """
        null_positions = (UnknownPosition, NoPosition)
        if self.ring_start in null_positions or self.ring_end in null_positions:
            return RingType.x
        try:
            diff = self.ring_end - self.ring_start
            if diff == 4:
                return RingType.pyranose
            elif diff == 3:
                return RingType.furanose
            elif diff == 0:
                return RingType.open
        except TypeError:
            pass
        return RingType.x

    @property
    def reducing_end(self):
        """Return the reducing end type of
        `self` or |None| if `self` is not reduced. The reducing end value
        can also be found in :attr:`modifications`.

        If `Modification.aldi` is present, it will be converted into
        an instance of :class:`.ReducedEnd` with default arguments.

        TODO: Remove Redundancy between `aldi` check and reduction.

        Returns
        -------
        ReducedEnd or None
        """
        if self._reducing_end is None:
            if not self._checked_for_reduction:
                for pos, mod in list(self.modifications.items()):
                    if isinstance(mod, ReducedEnd):
                        self._reducing_end = mod
                        break
                    elif mod is Modification.aldi:  # pragma: no cover
                        self.modifications.popv(Modification.aldi)
                        self.reducing_end = ReducedEnd()
                        break
            else:
                return self._reducing_end
        if self._reducing_end is None:
            self._checked_for_reduction = True
        return self._reducing_end

    @reducing_end.setter
    def reducing_end(self, value):
        """Sets the reducing end type of `self`.

        If `value` is |None|, then this residue is not reduced.
        Else, if `value` ``is True``, then this residue's reducing end
        is set to an instance of :class:`.ReducedEnd` with default parameters.
        Else, this residue's reducing_end is set to `value`.

        This process will scan :attr:`modifications` for a pre-existing `reducing_end`
        value to replace.

        Parameters
        ----------
        value: True, None, or ReducedEnd

        """
        red_end = self.reducing_end
        if red_end is not None:
            self.modifications.pop(1, red_end)
        # Setting to None will un-reduce the sugar, but should not
        # put a None in :attr:`modifications`
        self._checked_for_reduction = False
        if value:
            if value is True:
                value = ReducedEnd()
            self.modifications[1] = value
            self._reducing_end = value
        else:
            self._reducing_end = value

    def _fast_reduce(self, value):
        """Expedite adding a reducing end to this monosaccharide. Assumes
        that `value` is a :class:`ReducedEnd` and that this monosaccharide is not
        already reduced.

        Parameters
        ----------
        value : ReducedEnd
            Description
        """
        if value is not None:
            self.modifications[1] = value
            self._reducing_end = value

    def __getitem__(self, position):
        '''
        Gets the collection of alterations made to the carbohydrate
        backbone at `position`. This queries :attr:`modifications`, :attr:`links`, and
        :attr:`substituent_links`.

        Returns
        -------
        :class:`dict`
        '''
        return {
            'modifications': self.modifications[position],
            'substituent_links': self.substituent_links[position],
            'links': self.links[position]
        }

    def _is_full(self, max_occupancy=0):
        return self._remaining_capacity() >= 0

    def _remaining_capacity(self):
        bonds = self.degree(deep=True)
        max_size = self.superclass.value - 1
        if self.ring_type == RingType.open:
            max_size += 1
        return max_size - bonds

    def open_attachment_sites(self, max_occupancy=0):
        '''
        When attaching :class:`Monosaccharide` instances to other objects,
        bonds are formed between the carbohydrate backbone and the other object.
        If a site is already bound, the occupying object fills that space on the
        backbone and prevents other objects from binding there.

        Currently only cares about the availability of the hydroxyl group. As there
        is not a hydroxyl attached to the ring-ending carbon, that should not be
        considered an open site.

        If any existing attached units have unknown positions, we can't provide any
        known positions, in which case the list of open positions will be a :class:`list`
        of ``-1`` s of the length of open sites.

        Parameters
        ----------
        max_occupancy: int
            The number of objects that may already be bound at a site before it
            is considered unavailable for attachment.

        Returns
        -------
        :class:`list`:
            The positions open for binding
        :class:`int`:
            The number of bound but unknown locations on the backbone.
        '''
        slots, unknowns = self._backbone_occupancy_list()

        null_positions = [
            UnknownPosition, 'x', NoPosition
        ]
        open_slots = []
        can_determine_positions = unknowns > 0 or (self.ring_end in null_positions)

        for i in range(len(slots)):
            if slots[i] <= max_occupancy and (i + 1) != self.ring_end:
                open_slots.append(
                    (i + 1) if not can_determine_positions else UnknownPosition)

        if self.ring_end in null_positions:
            open_slots.pop()

        return open_slots, unknowns

    def _backbone_occupancy_list(self):
        slots = [0] * self.superclass.value
        unknowns = 0
        null_positions = (UnknownPosition, 'x')
        for pos, obj in chain(self.modifications.items(),
                              self.links.items(),
                              self.substituent_links.items()):
            if obj == Modification.keto:
                continue
            if pos in null_positions:  # pragma: no cover
                unknowns += 1
            else:
                slots[pos - 1] += 1
        return slots, unknowns

    def occupied_attachment_sites(self):
        slots, unknowns = self._backbone_occupancy_list()
        occupied_sites = []
        for i, occupied in enumerate(slots, start=1):
            if occupied:
                occupied_sites.append(i)
        if self.ring_end != UnknownPosition:
            occupied_sites.append(self.ring_end)
        for i in range(unknowns):
            occupied_sites.append(UnknownPosition)
        return occupied_sites

    def total_attachement_sites(self):
        slots = self.superclass.value
        if self.ring_type != RingType.open:
            slots -= 1
        return slots

    def is_occupied(self, position):
        '''
        Checks to see if a particular backbone position is occupied by a :class:`~.constants.Modification`,
        :class:`~.substituent.Substituent`, or :class:`~.link.Link` to another :class:`Monosaccharide`.

        Parameters
        ----------
        position: int
            The position to check for occupancy. Passing -1 checks for undetermined attachments.

        Returns
        -------
        int:
            The number of occupants at ``position``

        Raises
        ------
        IndexError:
            When the position is less than 1 or exceeds the limits of the carbohydrate backbone's size.

        '''

        null_positions = (UnknownPosition, NoPosition, 'x')

        if (position > self.superclass.value) or (position < 1 and position not in null_positions):
            raise IndexError("Index out of bounds")
        # The unknown position is always available
        if position in (UnknownPosition, 'x'):  # pragma: no cover
            return 0
        n_occupants = len(self.links[position]) +\
            len(self.modifications[position]) +\
            len(self.substituent_links[position])
        if Modification.keto in self.modifications[position]:
            n_occupants -= 1
        return n_occupants

    def add_modification(self, modification, position, max_occupancy=0):
        '''
        Adds a modification instance to :attr:`modifications` at the site given
        by ``position``. This directly modifies :attr:`composition`, consequently
        changing :meth:`mass`

        Parameters
        ----------
        modification: str or Modification
            The modification to add. If passed a |str|, it will be
            translated into an instance of :class:`~glypy.structure.constants.Modification`
        position: int or 'x'
            The location to add the :class:`~.constants.Modification` to.
        max_occupancy: int, optional
            The maximum number of items acceptable at `position`. defaults to :const:`1`

        Raises
        ------
        IndexError
            ``position`` exceeds the bounds set by :attr:`superclass`.
        ValueError
            ``position`` is occupied by more than ``max_occupancy`` elements

        Returns
        -------
        :class:`Monosaccharide`:
            `self`, for chain calls
        '''
        if self.is_occupied(position) > max_occupancy:
            raise ValueError("Site is already occupied")
        self._checked_for_reduction = False
        is_keto = modification == Modification.keto
        if modification is Modification.aldi:  # pragma: no cover
            self.reducing_end = ReducedEnd()
        else:
            try:
                self.composition += modification_compositions[modification](position)
                self.modifications[position] = Modification[modification]
            # OO Modification
            except Exception:  # pragma: no cover
                self.composition += modification.composition
                self.modifications[position] = modification
        if is_keto:
            n_ketos = 0
            for position, modification in self.modifications.items():
                if modification == Modification.keto:
                    n_ketos += 1
            if n_ketos == 1:
                self.composition += Composition("H2")
        return self

    def drop_modification(self, position, modification):
        '''
        Remove the `modification` at `position`

        Parameters
        ----------
        position: int
            The position to drop the modification from
        modification: Modification
            The Modification to remove.

        Raises
        ------
        IndexError:
            If `position` is not a valid carbohydrate backbone position

        ValueError:
            If `modification` is not found at `position`

        Returns
        -------
        :class:`Monosaccharide`:
            `self`, for chain calls
        '''
        if position > self.superclass.value or position < 0 and position not in (UnknownPosition, 'x'):
            raise IndexError("Index out of bounds")
        try:
            self.modifications.pop(position, modification)
        except IndexError:
            raise ValueError("Modification {} not found at {}".format(modification, position))
        is_keto = modification == Modification.keto
        self._checked_for_reduction = False
        if modification is Modification.aldi:  # pragma: no cover
            self.reducing_end = None
        else:
            try:
                self.composition = self.composition - \
                    modification_compositions[modification](position)
            # OO Modification
            except Exception:  # pragma: no cover
                self.composition = self.composition - modification.composition
        if is_keto:
            n_ketos = 0
            for position, modification in self.modifications.items():
                if modification == Modification.keto:
                    n_ketos += 1
            if n_ketos == 0:
                self.composition -= Composition("H2")
        return self

    def add_substituent(self, substituent, position=-1, max_occupancy=0,
                        child_position=1, parent_loss=None, child_loss=None):
        '''
        Adds a :class:`~glypy.structure.substituent.Substituent` and associated :class:`~glypy.structure.link.Link`
        to :attr:`substituent_links` at the site given by ``position``. This new substituent is included when
        calculating mass with substituents included.

        >>> from glypy import monosaccharides
        >>> hex = monosaccharides.Hex
        >>> hexnac = monosaccharides.HexNAc
        >>> hex.add_substituent("n-acetyl", 2, parent_loss="OH")
        RES 1b:x-xx-HEX-1:5 2s:n-acetyl LIN 1:1d(2+1)2n
        >>> hexnac == hex
        True

        Parameters
        ----------
        substituent: str or Substituent
            The substituent to add. If passed a |str| it will be
            translated into an instance of |Substituent|.
        position: int or 'x'
            The location to add the |Substituent| link to :attr:`substituent_links`. Defaults to -1
        child_position: int
            The location to add the link to in `substituent` :attr:`links`. Defaults to -1. Substituent
            indices are currently not checked.
        max_occupancy: int, optional
            The maximum number of items acceptable at ``position``. Defaults to :const:`1`
        parent_loss: Composition or str
            The elemental composition removed from ``self``
        child_loss: Composition or str
            The elemental composition removed from ``substituent``

        Raises
        ------
        IndexError
            ``position`` exceeds the bounds set by :attr:`superclass`.
        ValueError
            ``position`` is occupied by more than ``max_occupancy`` elements


        Returns
        -------
        :class:`Monosaccharide`:
            `self`, for chain calls
        '''
        if self.is_occupied(position) > max_occupancy:
            raise ValueError("Site is already occupied")
        if isinstance(substituent, basestring):
            substituent = Substituent(substituent)
        if child_loss is None:
            child_loss = Composition("H")
        if parent_loss is None:
            parent_loss = substituent.attachment_composition_loss()
        Link(parent=self, child=substituent,
             parent_position=position, child_position=child_position,
             parent_loss=parent_loss, child_loss=child_loss)
        return self

    def drop_substituent(self, position, substituent=None, refund=True):
        '''
        Remove the `substituent` at `position`.

        If `substituent` is |None|, then the first substituent found at `position` is
        removed.

        >>> from glypy import monosaccharides
        >>> hex = monosaccharides.Hex
        >>> hexnac = monosaccharides.HexNAc
        >>> hexnac.drop_substituent(2)
        RES 1b:x-xx-HEX-1:5
        >>> hexnac == hex
        True

        Parameters
        ----------
        position: int
            The position to drop the modification from
        substituent: Substituent
            The |Substituent| to remove. If |None|, the first substituent found at
            `position` will be removed
        refund: bool
            Passed to :meth:`~.Link.break_link`

        Raises
        ------
        IndexError:
            If `position` is not a valid carbohydrate backbone position

        ValueError:
            If `substituent` is not found at `position`

        Returns
        -------
        :class:`Monosaccharide`:
            `self`, for chain calls
        '''
        if position > self.superclass.value or position < 0 and position not in (UnknownPosition, 'x'):
            raise IndexError("Index out of bounds")
        if isinstance(substituent, basestring):
            substituent = Substituent(substituent)
        link_obj = None
        for substituent_link in self.substituent_links[position]:
            if substituent is None or substituent_link.child.name == substituent.name:
                link_obj = substituent_link
                break
        if link_obj is None:
            if substituent is not None:
                msg = "No matching substituent found at {position}.".format(
                    position=position)
            else:
                msg = "No substituents found at {position}.".format(
                    position=position)
            raise IndexError(msg)

        link_obj.break_link(refund=refund)
        return self

    def add_monosaccharide(self, monosaccharide, position=-1, max_occupancy=0,
                           child_position=-1, parent_loss=None, child_loss=None):
        '''
        Adds a |Monosaccharide| and associated |Link| to :attr:`links` at the site given by
        ``position``.

        >>> from glypy import monosaccharides
        >>> hexnac = monosaccharides.HexNAc
        >>> hex = monosaccharides.Hex
        >>> hexnac.add_monosaccharide(hex, 1)
        RES 1b:x-xx-HEX-1:5 2s:n-acetyl LIN 1:1d(2+1)2n
        >>> hexnac.links[1][0].child
        RES 1b:x-xx-HEX-1:5

        Parameters
        ----------
        monosaccharide: Monosaccharide
            The monosaccharide to add.
        position: int or 'x'
            The location to add the |Monosaccharide| link to :attr:`links`. Defaults to -1
        child_position: int
            The location to add the link to in `monosaccharide`'s :attr:`links`. Defaults to -1.
        max_occupancy: int, optional
            The maximum number of items acceptable at ``position``. Defaults to :const:`1`
        parent_loss: Composition or str
            The elemental composition removed from ``self``
        child_loss: Composition or str
            The elemental composition removed from ``monosaccharide``

        Raises
        ------
        IndexError
            ``position`` exceeds the bounds set by :attr:`superclass`.
        ValueError
            ``position`` is occupied by more than ``max_occupancy`` elements


        Returns
        -------
        :class:`Monosaccharide`:
            `self`, for chain calls
        '''
        if parent_loss is None:
            parent_loss = Composition("H")
        if child_loss is None:
            child_loss = Composition("OH")
        if self.is_occupied(position) > max_occupancy:
            raise ValueError("Parent Site is already occupied")  # pragma: no cover
        if monosaccharide.is_occupied(child_position) > max_occupancy:
            raise ValueError("Child Site is already occupied")  # pragma: no cover
        Link(parent=self, child=monosaccharide,
             parent_position=position, child_position=child_position,
             parent_loss=parent_loss, child_loss=child_loss)
        return self

    def drop_monosaccharide(self, position, refund=True):
        '''
        Remove the glycosidic bond at `position`, detatching a connected |Monosaccharide|

        If there is more than one glycosidic bond at `position`, an error will be raised.

        >>> from glypy import glycans
        >>> n_linked_core = glycans["N-Linked Core"]
        >>> n_linked_core.root.drop_monosaccharide(4)
        RES 1b:b-dglc-HEX-1:5 2s:n-acetyl LIN 1:1d(2+1)2n
        >>> n_linked_core.mass()
        221.08993720321

        Parameters
        ----------
        position: int
            The position to drop the modification from
        refund: bool
            Passed to :meth:`~.Link.break_link`

        Raises
        ------
        IndexError:
            If `position` is not a valid carbohydrate backbone position

        ValueError:
            If no |Link| or more than one |Link| is found at `position`

        Returns
        -------
        :class:`Monosaccharide`:
            `self`, for chain calls
        '''
        if position > self.superclass.value or position < 0 and position not in (UnknownPosition, 'x'):
            raise IndexError("Index out of bounds")
        link_obj = None
        if len(self.links[position]) > 1:
            raise ValueError("Too many monosaccharides found")
        for link in self.links[position]:
            link_obj = link
            break
        if link_obj is None:
            raise ValueError(
                "No matching monosaccharide found at {position}".format(position=position))

        link_obj.break_link(refund=refund)
        return self

    def serialize(self, name='glycoct'):
        """Convert this object into text using the requested textual encoding

        Parameters
        ----------
        name: :class:`str`, optional
            The name of the textual encoding, e.g. "glycoct" or "iupac"

        Returns
        -------
        :class:`str`
        """
        return self._serializers[name.lower()](self)

    @classmethod
    def register_serializer(cls, name, method):
        """Add `method` as `name` to the set of serializers to pick from
        in :meth:`serialize`

        Parameters
        ----------
        name : str
            The name of the serializer
        method : Callable
            A callable object that when called with a :class:`Monosaccharide` returns a :class:`str`
        """
        cls._serializers[name.lower()] = method

    @classmethod
    def available_serializers(cls):
        """Get the list of available serialization formats

        Returns
        -------
        :class:`list` of :class:`str`
        """
        return list(cls._serializers.keys())

    def _flat_equality(self, other, lengths=True):
        '''
        Test for equality of all scalar-ish features that do not
        require recursively comparing links which in turn compare their
        connected units.
        '''
        llen = len
        flat = (self.anomer == other.anomer) and\
            (self.ring_start == other.ring_start) and\
            (self.ring_end == other.ring_end) and\
            (self.superclass == other.superclass) and\
            (self.modifications) == (other.modifications) and\
            (self.configuration) == (other.configuration) and\
            (self.stem) == (other.stem)
        if lengths:
            flat = flat and\
                llen(self.links) == llen(other.links) and\
                llen(self.substituent_links) == llen(other.substituent_links) and\
                self.total_composition() == other.total_composition()
        return flat

    def exact_ordering_equality(self, other, substituents=True, visited=None):
        '''
        Performs equality testing between two monosaccharides where
        the exact position (and ordering by sort) of links must to match between
        the input |Monosaccharide| objects

        Returns
        -------
        |bool|
        '''
        if visited is None:
            visited = set()
        if (self.id, other.id) in visited:  # pragma: no cover
            return True

        visited.add((self.id, other.id))

        if self._flat_equality(other):
            if substituents:
                for a_sub, b_sub in zip(self.substituents(), other.substituents()):
                    if a_sub != b_sub:  # pragma: no cover
                        return False
            for a_mod, b_mod in zip(self.modifications.items(), other.modifications.items()):
                if a_mod != b_mod:  # pragma: no cover
                    return False
            for a_child, b_child in zip(self.children(True), other.children(True)):
                a_parent_pos, a_link = a_child
                b_parent_pos, b_link = b_child
                if (a_parent_pos != b_parent_pos) or (a_link.child_position != b_link.child_position):
                    return False
                if not a_link.child.exact_ordering_equality(b_link.child,
                                                            substituents=substituents,
                                                            visited=visited):  # pragma: no cover
                    return False
            return True
        return False

    def topological_equality(self, other, substituents=True, visited=None):
        '''
        Performs equality testing between two monosaccharides where
        the exact ordering of child links does not have to match between
        the input |Monosaccharide|s, so long as an exact match of the
        subtrees is found

        Returns
        -------
        |bool|
        '''
        if visited is None:
            visited = set()
        if (self.id, other.id) in visited:  # pragma: no cover
            return True
        if self._flat_equality(other) and (not substituents or self._match_substituents(other)):
            taken_b = set()
            b_children = list(other.children(links=True))
            a_children = list(self.children(links=True))
            for a_pos, a_link in a_children:
                a_child = a_link.child
                matched = False
                for b_pos, b_link in b_children:
                    b_child = b_link.child
                    if (b_pos, b_child.id) in taken_b:
                        continue
                    if a_child.topological_equality(b_child,
                                                    substituents=substituents,
                                                    visited=visited):
                        matched = True
                        taken_b.add((b_pos, b_child.id))
                        break
                if not matched and len(a_children) > 0:
                    return False
            if len(taken_b) != len(b_children):  # pragma: no cover
                return False
            return True
        return False

    def _match_substituents(self, other):
        '''
        Helper method for matching substituents in an order-independent
        fashion. Used by :meth:`topological_equality`
        '''
        taken_b = set()
        b_num_unknown = len(other.substituent_links[-1])
        unknown_cntr = 0
        b_substituents = list(other.substituents())
        cntr = 0
        for a_pos, a_substituent in self.substituents():
            matched = False
            cntr += 1
            for b_pos, b_substituent in b_substituents:
                if b_pos in taken_b:
                    if b_pos != UnknownPosition:  # pragma: no cover
                        continue
                    else:  # pragma: no cover
                        if unknown_cntr < b_num_unknown:
                            unknown_cntr += 1
                        else:
                            continue
                if b_substituent == a_substituent:
                    matched = True
                    taken_b.add(b_pos)
                    break
            if not matched and cntr > 0:
                return False
        if len(taken_b) + unknown_cntr != len(b_substituents):
            return False
        return True

    def __eq__(self, other):
        '''
        Test for equality between :class:`Monosaccharide` instances.
        First try scalar equality of fields, and then compare descendants.
        '''
        if (other is None):
            return False
        if not isinstance(other, Monosaccharide):
            return False
        return self.exact_ordering_equality(other)

    def __hash__(self):
        return hash(self.id)

    def __ne__(self, other):
        return not (self == other)

    __repr__ = serialize

    def __getstate__(self):
        state = dict()
        state['_anomer'] = self._anomer
        state['_superclass'] = self._superclass
        state['_stem'] = self._stem
        state['_configuration'] = self._configuration
        state['_degree'] = self._degree
        state['ring_start'] = self.ring_start
        state['ring_end'] = self.ring_end
        state['id'] = self.id
        state['modifications'] = self.modifications
        state['links'] = self.links
        state['substituent_links'] = self.substituent_links
        state['composition'] = self.composition
        state['_reducing_end'] = self._reducing_end
        return state

    def __setstate__(self, state):
        '''
        Does some testing to upgrade outdated, but equivalent
        modification models.
        '''
        self._checked_for_reduction = False
        self._anomer = state['_anomer']
        self._superclass = state['_superclass']
        self._stem = state['_stem']
        self._configuration = state['_configuration']
        self._degree = state.get("_degree")
        self.ring_start = state['ring_start']
        self.ring_end = state['ring_end']
        self.id = state['id']

        self.modifications = state['modifications']
        self.links = state['links']
        self.substituent_links = state['substituent_links']
        self.composition = state["composition"]
        reduced = state.get('_reducing_end', None)
        if self._degree is None:
            self._degree = len(self.links) + len(self.substituent_links)
        # Make sure that if "aldi" is present, to replace it with
        # the default ReducedEnd
        if reduced is None:
            if self.modifications.popv(Modification.aldi) is not None:  # pragma: no cover
                # Deduct the modification mass from the main composition
                self.composition -= {"H": 2}
                reduced = True
        self._reducing_end = None
        self.reducing_end = reduced

    def mass(self, average=False, charge=0, mass_data=None, substituents=True):
        '''
        Calculates the total mass of ``self``.

        Parameters
        ----------
        average: bool, optional, defaults to False
            Whether or not to use the average isotopic composition when calculating masses.
            When ``average == False``, masses are calculated using monoisotopic mass.
        charge: int, optional, defaults to 0
            If charge is non-zero, m/z is calculated, where m is the theoretical mass, and z is ``charge``
        mass_data: dict, optional
            If mass_data is None, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of mass_data are assumed to contain elemental mass and isotopic abundance information.
            Defaults to :const:`None`.
        substituents: bool, optional, defaults to True
            Whether or not to include substituents' masses.
        Returns
        -------
        :class:`float`

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        '''
        if charge == 0:
            mass = calculate_mass(
                self.composition, average=average, charge=0, mass_data=mass_data)
            if substituents:
                subst_visited = set()
                for substituent_link in self.substituent_links.values():
                    subst = substituent_link[self]
                    if subst.id in subst_visited:
                        continue
                    subst_visited.add(subst.id)
                    mass += subst.mass(
                        average=average, charge=0, mass_data=mass_data)
            if self._reducing_end is not None:
                mass += self._reducing_end.mass(
                    average=average, charge=0, mass_data=mass_data)
        else:
            mass = self.total_composition().calc_mass(average=average, charge=charge, mass_data=mass_data)
        return mass

    def total_composition(self):
        '''
        Computes the sum of the composition of ``self`` and each of its linked
        :class:`~glypy.structure.substituent.Substituent`\ s

        Returns
        -------
        :class:`~glypy.composition.Composition`
        '''
        comp = self.composition.clone()
        for pos, sub in self.substituents():
            comp += sub.total_composition()
        red_end = self.reducing_end
        if red_end is not None:
            comp += red_end.total_composition()
        return comp

    def children(self, links=False):
        '''
        Returns an iterator over the :class:`Monosaccharide` instancess which are considered
        the descendants of `self`

        >>> from glypy import glycans
        >>> n_linked_core = glycans["N-Linked Core"]
        >>> ch = n_linked_core.root.children()
        >>> ch[0]
        (4, RES 1b:b-dglc-HEX-1:5 2s:n-acetyl LIN 1:1d(2+1)2n)
        >>>

        Parameters
        ----------
        links: bool
            Whether to return the Link objects, or their children.
            Defaults to False

        Returns
        -------
        :class:`list` of
        position: int
            Location of the bond to the child |Monosaccharide|
        child: Monosaccharide
            |Monosaccharide| at `position`
        '''
        if links:
            result = [(pos, link) for pos, link in self.links.items() if not link.is_child(self)]
        else:
            result = [(pos, link.child) for pos, link in self.links.items() if not link.is_child(self)]
        return result

    def parents(self, links=False):
        '''
        Returns an iterator over the :class:`Monosaccharide` instances which are considered
        the ancestors of `self`.

        links: bool
            Whether to return the Link objects, or their parents.
            Defaults to False

        Returns
        -------
        :class:`list` of
        position: int
            Location of the bond to the parent |Monosaccharide|
        parent: Monosaccharide
            |Monosaccharide| at `position`
        '''
        if links:
            result = [(pos, link) for pos, link in self.links.items() if not link.is_parent(self)]
        else:
            result = [(pos, link.parent) for pos, link in self.links.items() if not link.is_parent(self)]
        return result

    def substituents(self):
        '''
        Returns an iterator over all substituents attached
        to :obj:`self` by a :class:`~.link.Link` object stored in :attr:`~.substituent_links`

        Returns
        -------
        :class:`list` of
        position: int
            Location of the bond to the substituent
        substituent: Substituent
            |Substituent| at `position`
        '''
        subst_visited = set()
        results = []
        for pos, link in self.substituent_links.items():
            subst = link.to(self)
            if subst.id in subst_visited:
                continue
            subst_visited.add(subst.id)
            results.append((pos, subst))
        return results

    def degree(self, deep=False):
        '''
        Return the "graph theory" degree of this molecule

        Returns
        -------
        int
        '''
        if deep:
            res = len(self.links) + len(self.substituent_links)
            if hasattr(self, "_degree"):
                assert res == self._degree
        return self._degree

    def __iter__(self):
        return iter(self.children())

    def has_undefined_linkages(self):
        for link in self.links.values():
            if link.parent_position == UnknownPosition or link.child_position == UnknownPosition:
                return True
        for link in self.substituent_links.values():
            if link.parent_position == UnknownPosition or link.child_position == UnknownPosition:
                return True
        return False


class ReducedEnd(object):
    """Represents the composition shift and conformation change created
    by reducing a |Monosaccharide|.

    Attributes
    ----------
    composition : |Composition|
        The elemental composition of the reducing end reduction modification.
    links: |OrderedMultiMap|
        The attached substituents
    valence: |int|
        Number of substituents this node can host
    id: |int|
        Unique identifier

    There is also a class attribute, :attr:`name` for comparison with :attr:`~.Modification.aldi`
    """

    node_type = object()
    name = 'aldi'

    def __init__(self, composition=None, substituents=None, valence=1, id=None):
        if composition is None:
            composition = Composition("H2")
        else:
            composition = Composition(composition)
        self.composition = composition
        self.base_composition = self.composition.clone()
        self.links = substituents or OrderedMultiMap()
        self.valence = valence
        self.id = id or uid()
        self._degree = len(self.links)

    def is_occupied(self, position):
        '''
        Checks to see if a particular backbone position is occupied by a or :class:`.Substituent`.

        Parameters
        ----------
        position: int
            The position to check for occupancy. Passing -1 checks for undetermined attachments.

        Returns
        -------
        numeric:
            The number of occupants at ``position``, or `float('inf')` if `position` exceeds :attr:`valence`
        '''
        if position > self.valence:  # pragma: no cover
            return float('inf')
        else:
            return len(self.links[position])

    def add_substituent(self, substituent, position=-1, max_occupancy=0,
                        child_position=1, parent_loss=None, child_loss=None):
        '''
        Adds a |Substituent| and associated |Link|
        to :attr:`substituent_links` at the site given by ``position``. This new substituent is included when
        calculating mass with substituents included

        Parameters
        ----------
        substituent: str or Substituent
            The substituent to add. If passed a |str|, it will be
            translated into an instance of |Substituent|
        position: int or 'x'
            The location to add the |Substituent| link to :attr:`substituent_links`. Defaults to -1
        child_position: int
            The location to add the link to in `substituent`'s :attr:`links`. Defaults to -1. Substituent
            indices are currently not checked.
        max_occupancy: int, optional
            The maximum number of items acceptable at ``position``. Defaults to :const:`1`
        parent_loss: Composition or str
            The elemental composition removed from ``self``
        child_loss: Composition or str
            The elemental composition removed from ``substituent``

        Raises
        ------
        IndexError
            ``position`` exceeds the bounds set by :attr:`superclass`.
        ValueError
            ``position`` is occupied by more than ``max_occupancy`` elements
        '''
        if self.is_occupied(position) > max_occupancy:  # pragma: no cover
            raise ValueError("Site is already occupied")
        if child_loss is None:
            child_loss = Composition("H")
        if parent_loss is None:
            parent_loss = Composition("H")
        if isinstance(substituent, basestring):
            substituent = Substituent(substituent)
        Link(parent=self, child=substituent,
             parent_position=position, child_position=child_position,
             parent_loss=parent_loss, child_loss=child_loss)
        return self

    def drop_substituent(self, position, substituent=None, refund=True):
        '''
        Remove the `substituent` at `position`.

        If `substituent` is |None|, then the first substituent found at `position` is
        removed.

        Parameters
        ----------
        position: int
            The position to drop the modification from
        substituent: Substituent
            The |Substituent| to remove. If |None|, the first substituent found at
            `position` will be removed
        refund: bool
            Passed to :meth:`~.Link.break_link`

        Raises
        ------
        IndexError:
            If `position` exceeds :attr:`valence`

        ValueError:
            If `substituent` is not found at `position`

        Returns
        -------
        ReducedEnd:
            `self` for chaining calls
        '''
        if position > self.valence:  # pragma: no cover
            raise IndexError("Index out of bounds")
        if isinstance(substituent, basestring):
            substituent = Substituent(substituent)
        link_obj = None
        for substituent_link in self.links[position]:
            if substituent_link.child.name == substituent.name or substituent is None:
                link_obj = substituent_link
                break
        if link_obj is None:
            if substituent is not None:
                msg = "No matching substituent found at {position}.".format(
                    position=position)
            else:
                msg = "No substituents found at {position}.".format(
                    position=position)
            raise IndexError(msg)

        link_obj.break_link(refund=refund)
        return self

    def mass(self, average=False, charge=0, mass_data=None):
        '''
        Calculates the total mass of ``self``.

        Parameters
        ----------
        average: bool, optional, defaults to False
            Whether or not to use the average isotopic composition when calculating masses.
            When ``average == False``, masses are calculated using monoisotopic mass.
        charge: int, optional, defaults to 0
            If charge is non-zero, m/z is calculated, where m is the theoretical mass, and z is ``charge``
        mass_data: dict, optional
            If mass_data is None, standard NIST mass and isotopic abundance data are used. Otherwise the
            contents of mass_data are assumed to contain elemental mass and isotopic abundance information.
            Defaults to :const:`None`.

        Returns
        -------
        :class:`float`

        See also
        --------
        :func:`glypy.composition.composition.calculate_mass`
        '''
        mass = calculate_mass(
            self.composition, average=average, charge=charge, mass_data=mass_data)

        for pos, link in self.links.items():
            mass += link[self].mass(average=average, charge=charge, mass_data=mass_data)
        return mass

    def clone(self, prop_id=True):
        """Make a deep copy of `self`.

        Parameters
        ----------
        prop_id: bool
            Whether to copy over :attr:`id`.

        Returns
        -------
        ReducedEnd
        """
        result = ReducedEnd(
            Composition(self.base_composition), substituents=None,
            valence=self.valence, id=self.id if prop_id else None)
        for position, link in self.links.items():
            result.add_substituent(link.child.clone(), position)
        return result

    def children(self):
        '''
        Returns an iterator over the nodes which are considered
        the descendants of ``self``.
        '''
        for pos, link in self.links.items():
            if link.is_child(self):
                continue
            yield (pos, link.child)

    def total_composition(self):
        '''
        Computes the sum of the composition of `self` and each of its linked
        :class:`~.substituent.Substituent`s

        Returns
        -------
        :class:`~glypy.composition.Composition`
        '''
        comp = self.composition
        for pos, sub in self.children():
            comp = comp + sub.total_composition()
        return comp

    def __repr__(self):  # pragma: no cover
        rep = "<ReducedEnd {}>".format(self.total_composition())
        return rep

    def __eq__(self, other):
        '''
        Test for equality with `other`, with special handling for `EnumValue` comparisons
        for dealing with legacy format objects. If compared to the string "aldi", returns |True|,
        otherwise does deep comparison of composition and substituents.
        '''
        if other is None:
            return False
        elif isinstance(other, (int, EnumValue)):
            return False
        elif isinstance(other, str):
            return other == self.name
        composition_equal = self.composition == other.composition
        if not composition_equal:
            return False
        for spair, opair in izip_longest(self.links.items(), other.links.items()):
            if spair[0] != opair[0]:  # pragma: no cover
                return False
            if not spair[1]._flat_equality(opair[1]):  # pragma: no cover
                return False
        return True

    def __hash__(self):
        return hash(self.name)

    def __ne__(self, other):
        return not self == other

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._degree = state.get("_degree", len(self.links))


MonosaccharideOccupancy = make_struct(
    "MonosaccharideOccupancy", ["id", "open_sites", "unknown_sites", "occupied_sites"])


def build_monosaccharide_occupancy(cls, monosaccharide):
    open_sites, unknown_sites = monosaccharide.open_attachment_sites()
    occupied_sites = monosaccharide.occupied_attachment_sites()
    return cls(monosaccharide.id, open_sites, unknown_sites, occupied_sites)


MonosaccharideOccupancy.build = classmethod(
    build_monosaccharide_occupancy)


class AnnotatedMonosaccharide(Monosaccharide):
    __slots__ = ('annotations', )

    def __init__(self, anomer=None, configuration=None, stem=None,
                 superclass=None, ring_start=UnknownPosition, ring_end=UnknownPosition,
                 modifications=None, links=None, substituent_links=None,
                 composition=None, reduced=None, id=None, fast=False, annotations=None):
        super(AnnotatedMonosaccharide, self).__init__(
            anomer, configuration, stem, superclass, ring_start, ring_end,
            modifications, links, substituent_links, composition, reduced,
            id, fast)
        if annotations is None:
            annotations = {}
        self.annotations = annotations

    def __getstate__(self):
        state = super(AnnotatedMonosaccharide, self).__getstate__()
        state['annotations'] = self.annotations
        return state

    def __setstate__(self, state):
        super(AnnotatedMonosaccharide, self).__setstate__(state)
        self.annotations = state.get("annotations", {})

    def clone(self, prop_id=False, fast=True, monosaccharide_type=None):
        '''
        Copies just this |Monosaccharide| and its |Substituent|s, creating a separate instance
        with the same data. All mutable data structures are duplicated and distinct from the original.

        Does not copy any :attr:`links` as this would cause recursive duplication of the entire |Glycan|
        graph.

        Returns
        -------
        Monosaccharide

        '''
        if monosaccharide_type is None:
            monosaccharide_type = self.__class__
        monosaccharide = super(AnnotatedMonosaccharide, self).clone(prop_id, fast, monosaccharide_type)
        monosaccharide.annotations = self.annotations.copy()
        return monosaccharide
